import mock

from uaclient import apt
from uaclient.api.u.security.package_manifest.v1 import _package_manifest
from uaclient.snap import SnapPackage

M_PATH = "uaclient.api.u.security.package_manifest.v1"


@mock.patch("uaclient.snap.get_snap_info")
@mock.patch("uaclient.snap.system.subp")
@mock.patch(M_PATH + ".apt.get_installed_packages")
class TestPackageInstalledV1:
    def test_snap_packages_added(
        self, m_installed_apt_pkgs, m_sys_subp, m_get_snap_info, FakeConfig
    ):
        m_installed_apt_pkgs.return_value = []
        m_sys_subp.return_value = (
            "Name  Version Rev Tracking Publisher Notes\n"
            "helloworld 6.0.16 126 latest/stable dev1 -\n"
            "bare 1.0 5 latest/stable canonical** base\n"
            "canonical-livepatch 10.2.3 146 latest/stable canonical** -\n"
        ), ""
        m_get_snap_info.side_effect = [
            SnapPackage(
                "helloworld",
                "6.0.16",
                "126",
                "latest/stable",
                "dev1",
            ),
            SnapPackage(
                "bare",
                "1.0",
                "5",
                "latest/stable",
                "canonical**",
            ),
            SnapPackage(
                "canonical-livepatch",
                "10.2.3",
                "146",
                "latest/stable",
                "canonical**",
            ),
        ]
        result = _package_manifest(FakeConfig())
        assert (
            "snap:helloworld\tlatest/stable\t126\n"
            + "snap:bare\tlatest/stable\t5\n"
            + "snap:canonical-livepatch\tlatest/stable\t146\n"
            == result.manifest_data
        )

    def test_apt_packages_added(
        self, m_installed_apt_pkgs, m_sys_subp, m_get_snap_info, FakeConfig
    ):
        m_sys_subp.return_value = "", ""
        apt_pkgs = [
            apt.InstalledAptPackage(name="one", arch="all", version="4:1.0.2"),
            apt.InstalledAptPackage(name="two", arch="amd64", version="0.1.1"),
        ]
        m_installed_apt_pkgs.return_value = apt_pkgs
        result = _package_manifest(FakeConfig())
        assert "one\t4:1.0.2\ntwo:amd64\t0.1.1\n" == result.manifest_data

    def test_apt_snap_packages_added(
        self, m_installed_apt_pkgs, m_sys_subp, m_get_snap_info, FakeConfig
    ):
        apt_pkgs = [
            apt.InstalledAptPackage(name="one", arch="all", version="4:1.0.2"),
            apt.InstalledAptPackage(name="two", arch="amd64", version="0.1.1"),
        ]
        m_sys_subp.return_value = (
            "Name  Version Rev Tracking Publisher Notes\n"
            "helloworld 6.0.16 126 latest/stable dev1 -\n"
            "bare 1.0 5 latest/stable canonical** base\n"
            "canonical-livepatch 10.2.3 146 latest/stable canonical** -\n"
        ), ""
        m_get_snap_info.side_effect = [
            SnapPackage(
                "helloworld",
                "6.0.16",
                "126",
                "latest/stable",
                "dev1",
            ),
            SnapPackage(
                "bare",
                "1.0",
                "5",
                "latest/stable",
                "canonical**",
            ),
            SnapPackage(
                "canonical-livepatch",
                "10.2.3",
                "146",
                "latest/stable",
                "canonical**",
            ),
        ]
        m_installed_apt_pkgs.return_value = apt_pkgs
        result = _package_manifest(FakeConfig())
        assert (
            "one\t4:1.0.2\ntwo:amd64\t0.1.1\n"
            + "snap:helloworld\tlatest/stable\t126\n"
            + "snap:bare\tlatest/stable\t5\n"
            + "snap:canonical-livepatch\tlatest/stable\t146\n"
            == result.manifest_data
        )



from enum import Enum
from typing import List, Optional

from uaclient import exceptions, livepatch
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    Field,
    StringDataValue,
    data_list,
)
from uaclient.system import (
    get_kernel_info,
    get_reboot_required_pkgs,
    should_reboot,
)


class RebootRequiredPkgs(DataObject):
    fields = [
        Field(
            "standard_packages",
            data_list(StringDataValue),
            False,
            doc="Non-kernel packages that require a reboot",
        ),
        Field(
            "kernel_packages",
            data_list(StringDataValue),
            False,
            doc="Kernel packages that require a reboot",
        ),
    ]

    def __init__(
        self,
        standard_packages: Optional[List[str]],
        kernel_packages: Optional[List[str]],
    ):
        self.standard_packages = standard_packages
        self.kernel_packages = kernel_packages


class RebootRequiredResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "reboot_required",
            StringDataValue,
            doc="Either 'yes', 'no', or 'yes-kernel-livepatches-applied'",
        ),
        Field(
            "reboot_required_packages",
            RebootRequiredPkgs,
            doc="The packages that require a reboot",
        ),
        Field(
            "livepatch_enabled_and_kernel_patched",
            BoolDataValue,
            doc="True if livepatch is enabled and working",
        ),
        Field(
            "livepatch_enabled",
            BoolDataValue,
            doc="True if livepatch is enabled",
        ),
        Field(
            "livepatch_state",
            StringDataValue,
            False,
            doc="The state of livepatch as reported by the livepatch client",
        ),
        Field(
            "livepatch_support",
            StringDataValue,
            False,
            doc="Whether livepatch covers the current kernel",
        ),
    ]

    def __init__(
        self,
        reboot_required: str,
        reboot_required_packages: RebootRequiredPkgs,
        livepatch_enabled_and_kernel_patched: bool,
        livepatch_enabled: bool,
        livepatch_state: Optional[str],
        livepatch_support: Optional[str],
    ):
        self.reboot_required = reboot_required
        self.reboot_required_packages = reboot_required_packages
        self.livepatch_enabled_and_kernel_patched = (
            livepatch_enabled_and_kernel_patched
        )
        self.livepatch_enabled = livepatch_enabled
        self.livepatch_state = livepatch_state
        self.livepatch_support = livepatch_support


class RebootStatus(Enum):
    REBOOT_REQUIRED = "yes"
    REBOOT_NOT_REQUIRED = "no"
    REBOOT_REQUIRED_LIVEPATCH_APPLIED = "yes-kernel-livepatches-applied"


def _get_reboot_status():
    if not should_reboot():
        return RebootStatus.REBOOT_NOT_REQUIRED

    reboot_required_pkgs = get_reboot_required_pkgs()

    if not reboot_required_pkgs:
        return RebootStatus.REBOOT_REQUIRED

    # We will only check the Livepatch state if all the
    # packages that require a reboot are kernel related
    if reboot_required_pkgs.standard_packages:
        return RebootStatus.REBOOT_REQUIRED

    # If there are no kernel packages to cover or livepatch is not installed,
    # we should only return that a reboot is required
    if (
        not reboot_required_pkgs.kernel_packages
        or not livepatch.is_livepatch_installed()
    ):
        return RebootStatus.REBOOT_REQUIRED

    our_kernel_version = get_kernel_info().proc_version_signature_version

    try:
        lp_status = livepatch.status()
    except exceptions.ProcessExecutionError:
        return RebootStatus.REBOOT_REQUIRED

    if (
        lp_status is not None
        and our_kernel_version is not None
        and our_kernel_version == lp_status.kernel
        and lp_status.livepatch is not None
        and (
            lp_status.livepatch.state == "applied"
            or lp_status.livepatch.state == "nothing-to-apply"
        )
        and lp_status.supported == "supported"
    ):
        return RebootStatus.REBOOT_REQUIRED_LIVEPATCH_APPLIED

    # Any other Livepatch status will not be considered here to modify the
    # reboot state
    return RebootStatus.REBOOT_REQUIRED


def reboot_required() -> RebootRequiredResult:
    return _reboot_required(UAConfig())


def _reboot_required(cfg: UAConfig) -> RebootRequiredResult:
    """
    This endpoint informs if the system should be rebooted or not. Possible
    outputs are:

    #. ``yes``: The system should be rebooted.
    #. ``no``: There is no known need to reboot the system.
    #. ``yes-kernel-livepatches-applied``: There are Livepatch patches applied
       to the current kernel, but a reboot is required for an update to take
       place. This reboot can wait until the next maintenance window.
    """
    reboot_status = _get_reboot_status()
    reboot_required_pkgs = get_reboot_required_pkgs()
    livepatch_status = livepatch.status()

    if not livepatch_status:
        livepatch_enabled_and_kernel_patched = False
        livepatch_enabled = False
        livepatch_state = None
        livepatch_support = None
    else:
        livepatch_enabled = True
        livepatch_support = livepatch_status.supported
        livepatch_state = (
            livepatch_status.livepatch.state
            if livepatch_status.livepatch
            else None
        )
        if (
            livepatch_state not in ("applied", "nothing-to-apply")
            and livepatch_support != "supported"
        ):
            livepatch_enabled_and_kernel_patched = False
        else:
            livepatch_enabled_and_kernel_patched = True

    return RebootRequiredResult(
        reboot_required=reboot_status.value,
        reboot_required_packages=RebootRequiredPkgs(
            standard_packages=(
                reboot_required_pkgs.standard_packages
                if reboot_required_pkgs
                else None
            ),
            kernel_packages=(
                reboot_required_pkgs.kernel_packages
                if reboot_required_pkgs
                else None
            ),
        ),
        livepatch_enabled_and_kernel_patched=livepatch_enabled_and_kernel_patched,  # noqa
        livepatch_enabled=livepatch_enabled,
        livepatch_state=livepatch_state,
        livepatch_support=livepatch_support,
    )


endpoint = APIEndpoint(
    version="v1",
    name="RebootRequired",
    fn=_reboot_required,
    options_cls=None,
)

_doc = {
    "introduced_in": "27.12",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.pro.security.status.reboot_required.v1 import reboot_required

result = reboot_required()
""",  # noqa: E501
    "result_class": RebootRequiredResult,
    "exceptions": [],
    "example_cli": "pro api u.pro.security.status.reboot_required.v1",
    "example_json": """
{
    "reboot_required": "yes|no|yes-kernel-livepatches-applied"
}
""",
}



import mock
import pytest

from uaclient.cli import main
from uaclient.cli.config import unset_subcommand
from uaclient.entitlements.entitlement_status import ApplicationStatus
from uaclient.exceptions import NonRootUserError

M_LIVEPATCH = "uaclient.entitlements.livepatch."


@mock.patch("uaclient.log.setup_cli_logging")
@mock.patch("uaclient.contract.get_available_resources")
class TestMainConfigUnSet:
    @pytest.mark.parametrize(
        "kv_pair,err_msg",
        (
            (
                "junk",
                "<key> must be one of: http_proxy, https_proxy,"
                " apt_http_proxy, apt_https_proxy, ua_apt_http_proxy,"
                " ua_apt_https_proxy, global_apt_http_proxy,"
                " global_apt_https_proxy, update_messaging_timer,"
                " metering_timer",
            ),
            (
                "http_proxys",
                "<key> must be one of: http_proxy, https_proxy,"
                " apt_http_proxy, apt_https_proxy, ua_apt_http_proxy,"
                " ua_apt_https_proxy, global_apt_http_proxy,"
                " global_apt_https_proxy, update_messaging_timer,"
                " metering_timer",
            ),
        ),
    )
    def test_set_error_with_help_on_invalid_key_value_pair(
        self,
        _m_resources,
        _logging,
        kv_pair,
        err_msg,
        capsys,
        FakeConfig,
        event,
    ):
        """Exit 1 and print help on invalid key_value_pair input param."""
        with pytest.raises(SystemExit):
            with mock.patch(
                "sys.argv", ["/usr/bin/ua", "config", "unset", kv_pair]
            ):
                with mock.patch(
                    "uaclient.config.UAConfig",
                    return_value=FakeConfig(),
                ):
                    main()
        out, err = capsys.readouterr()
        assert err_msg in err


@mock.patch("uaclient.config.user_config_file.user_config.write")
class TestActionConfigUnSet:
    @mock.patch("uaclient.util.we_are_currently_root", return_value=False)
    def test_set_error_on_non_root_user(
        self, we_are_currently_root, _write, FakeConfig
    ):
        """Root is required to run pro config unset."""
        args = mock.MagicMock(key="https_proxy")
        cfg = FakeConfig()
        with pytest.raises(NonRootUserError):
            unset_subcommand.action(args, cfg=cfg)

    @pytest.mark.parametrize(
        "key,livepatch_enabled",
        (
            ("http_proxy", False),
            ("https_proxy", False),
            ("http_proxy", True),
            ("https_proxy", True),
        ),
    )
    @mock.patch("uaclient.cli.config.unconfigure_livepatch_proxy")
    @mock.patch(M_LIVEPATCH + "LivepatchEntitlement.application_status")
    @mock.patch("uaclient.cli.config.unconfigure_snap_proxy")
    def test_set_http_proxy_and_https_proxy_affects_snap_and_maybe_livepatch(
        self,
        unconfigure_snap_proxy,
        livepatch_status,
        unconfigure_livepatch_proxy,
        _write,
        key,
        livepatch_enabled,
        FakeConfig,
    ):
        """Set updates snap and livepatch proxy, persist config and exits 0.

        Only update livepatch proxy config is livepatch is enabled.
        """
        if livepatch_enabled:
            livepatch_status.return_value = (
                ApplicationStatus.ENABLED,
                "",
            )
        else:
            livepatch_status.return_value = (
                ApplicationStatus.DISABLED,
                "",
            )
        args = mock.MagicMock(key=key)
        cfg = FakeConfig()
        unset_subcommand.action(args, cfg=cfg)
        assert [
            mock.call(protocol_type=key.split("_")[0])
        ] == unconfigure_snap_proxy.call_args_list
        if livepatch_enabled:
            assert [
                mock.call(protocol_type=key.split("_")[0])
            ] == unconfigure_livepatch_proxy.call_args_list
        else:
            assert [] == unconfigure_livepatch_proxy.call_args_list



import time

from behave import step

from features.steps.shell import when_i_run_command


@step("I start `{task}` command `{command}` in the background")
def start_task(context, task, command):
    when_i_run_command(
        context,
        f"systemd-run --no-block --unit={task} --property=Type=oneshot --property=RemainAfterExit=yes {command}",  # noqa: E501
        "with sudo",
    )


@step("I wait for the `{task}` command to complete")
def wait_for_task(context, task):
    while True:
        try:
            when_i_run_command(
                context, f"systemctl is-active {task}", "with sudo"
            )
            break
        except AssertionError:
            time.sleep(2)



from uaclient import apt, snap
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.data_types import DataObject, Field, StringDataValue


class PackageManifestResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "manifest_data",
            StringDataValue,
            doc=(
                "Manifest of ``apt`` and ``snap`` packages installed on the"
                " system"
            ),
        ),
    ]

    def __init__(self, manifest_data: str):
        self.manifest_data = manifest_data


# The return class was once called PackageManifestResults
# We are keeping compatibility
PackageManifestResults = PackageManifestResult


def package_manifest() -> PackageManifestResult:
    return _package_manifest(UAConfig())


def _package_manifest(cfg: UAConfig) -> PackageManifestResult:
    """
    This endpoint returns the status of installed packages (``apt`` and
    ``snap``), formatted as a manifest file (i.e., ``package_name\\tversion``).
    """
    manifest = ""
    apt_pkgs = apt.get_installed_packages()
    for apt_pkg in apt_pkgs:
        arch = "" if apt_pkg.arch == "all" else ":" + apt_pkg.arch
        manifest += "{}{}\t{}\n".format(apt_pkg.name, arch, apt_pkg.version)

    pkgs = snap.get_installed_snaps()
    for pkg in pkgs:
        manifest += "snap:{name}\t{channel}\t{revision}\n".format(
            name=pkg.name,
            channel=pkg.channel,
            revision=pkg.revision,
        )

    return PackageManifestResult(manifest_data=manifest)


endpoint = APIEndpoint(
    version="v1",
    name="Packages",
    fn=_package_manifest,
    options_cls=None,
)

_doc = {
    "introduced_in": "27.12",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.security.package_manifest.v1 import package_manifest

result = package_manifest()
""",
    "result_class": PackageManifestResult,
    "exceptions": [],
    "example_cli": "pro api u.security.package_manifest.v1",
    "example_json": """
{
    "package_manifest":"package1\\t1.0\\npackage2\\t2.3\\n"
}
""",
}



import copy
from collections import defaultdict

import mock
import pytest

from uaclient import exceptions, http
from uaclient.api.u.pro.security.fix._common import (
    API_V1_CVE_TMPL,
    API_V1_CVES,
    API_V1_NOTICE_TMPL,
    API_V1_NOTICES,
    CVE,
    USN,
    CVEPackageStatus,
    UASecurityClient,
    get_cve_affected_source_packages_status,
    get_related_usns,
    get_usn_affected_packages_status,
    merge_usn_released_binary_package_versions,
    override_usn_release_package_status,
    query_installed_source_pkg_versions,
)

M_PATH = "uaclient.api.u.pro.security.fix._common."

SAMPLE_GET_CVES_QUERY_PARAMS = {
    "query": "vq",
    "priority": "vpr",
    "package": "vpa",
    "limit": 1,
    "offset": 2,
    "component": "vc",
    "version": "vv",
    "status": "vs",
}

SAMPLE_GET_NOTICES_QUERY_PARAMS = {
    "cves": "cve",
    "release": "vq",
    "limit": 1,
    "offset": 2,
    "order": "vo",
}


CVE_ESM_PACKAGE_STATUS_RESPONSE = {
    "component": None,
    "description": "1.17-6ubuntu4.1+esm1",
    "pocket": "esm-infra",
    "release_codename": "focal",
    "status": "released",
}


SAMBA_CVE_STATUS_BIONIC = {
    "component": None,
    "description": "2:4.7.6+dfsg~ubuntu-0ubuntu2.19",
    "pocket": None,
    "release_codename": "bionic",
    "status": "released",
}
SAMBA_CVE_STATUS_FOCAL = {
    "component": None,
    "description": "2:4.11.6+dfsg-0ubuntu1.4",
    "pocket": None,
    "release_codename": "focal",
    "status": "not-affected",
}
SAMBA_CVE_STATUS_UPSTREAM = {
    "component": None,
    "description": "",
    "pocket": None,
    "release_codename": "upstream",
    "status": "needs-triage",
}

SAMPLE_CVE_RESPONSE = {
    "bugs": ["https://bugzilla.samba.org/show_bug.cgi?id=14497"],
    "description": "\nAn elevation of privilege vulnerability exists ...",
    "id": "CVE-2020-1472",
    "notes": [{"author": "..", "note": "..."}],
    "notices_ids": ["USN-4510-1", "USN-4510-2", "USN-4559-1"],
    "packages": [
        {
            "debian": "https://tracker.debian.org/pkg/samba",
            "name": "samba",
            "source": "https://ubuntu.com/security/cve?package=samba",
            "statuses": [
                SAMBA_CVE_STATUS_BIONIC,
                SAMBA_CVE_STATUS_FOCAL,
                SAMBA_CVE_STATUS_UPSTREAM,
            ],
        }
    ],
    "status": "active",
}

SAMPLE_USN_RESPONSE = {
    "cves_ids": ["CVE-2020-1473", "CVE-2020-1472"],
    "id": "USN-4510-2",
    "instructions": "In general, a standard system update will make all ...\n",
    "references": [],
    "release_packages": {
        "series-example-1": [
            {
                "description": "SMB/CIFS file, print, and login ... Unix",
                "is_source": True,
                "name": "samba",
                "version": "2:4.3.11+dfsg-0ubuntu0.14.04.20+esm9",
            },
            {
                "is_source": False,
                "name": "samba",
                "source_link": "https://launchpad.net/ubuntu/+source/samba",
                "version": "2~14.04.1+esm9",
                "version_link": "https://....11+dfsg-0ubuntu0.14.04.20+esm9",
            },
        ],
        "series-example-2": [
            {
                "description": "high-level 3D graphics kit implementing ...",
                "is_source": True,
                "name": "coin3",
                "version": "3.1.4~abc9f50-4ubuntu2+esm1",
            },
            {
                "is_source": False,
                "name": "libcoin80-runtime",
                "source_link": "https://launchpad.net/ubuntu/+source/coin3",
                "version": "3~18.04.1+esm2",
                "version_link": "https://coin3...18.04.1+esm2",
            },
        ],
    },
    "summary": "Samba would allow unintended access to files over the ....\n",
    "title": "Samba vulnerability",
    "type": "USN",
}


SAMPLE_USN_RESPONSE_NO_CVES = {
    "cves_ids": [],
    "id": "USN-4038-3",
    "instructions": "In general, a standard system update will make all ...\n",
    "references": ["https://launchpad.net/bugs/1834494"],
    "release_packages": {
        "bionic": [
            {
                "description": "high-level 3D graphics kit implementing ...",
                "is_source": True,
                "name": "coin3",
                "version": "3.1.4~abc9f50-4ubuntu2+esm1",
            },
            {
                "is_source": False,
                "name": "libcoin80-runtime",
                "source_link": "https://launchpad.net/ubuntu/+source/coin3",
                "version": "3~18.04.1+esm2",
                "version_link": "https://coin3...18.04.1+esm2",
                "pocket": "security",
            },
        ]
    },
    "summary": "",
    "title": "USN vulnerability",
    "type": "USN",
}

CVE_PKG_STATUS_NEEDED = {
    "description": "2.1",
    "pocket": None,
    "status": "needed",
}
CVE_PKG_STATUS_IGNORED = {
    "description": "2.1",
    "pocket": None,
    "status": "ignored",
}
CVE_PKG_STATUS_DEFERRED = {
    "description": "2.1",
    "pocket": None,
    "status": "deferred",
}
CVE_PKG_STATUS_NEEDS_TRIAGE = {
    "description": "2.1",
    "pocket": None,
    "status": "needs-triage",
}
CVE_PKG_STATUS_PENDING = {
    "description": "2.1",
    "pocket": None,
    "status": "pending",
}
CVE_PKG_STATUS_RELEASED = {
    "description": "2.1",
    "pocket": "updates",
    "status": "released",
}
CVE_PKG_STATUS_RELEASED_ESM_INFRA = {
    "description": "2.1",
    "pocket": "esm-infra",
    "status": "released",
}
CVE_PKG_STATUS_RELEASED_ESM_APPS = {
    "description": "2.1",
    "pocket": "esm-apps",
    "status": "released",
}
CVE_PKG_STATUS_NEEDED = {"description": "", "pocket": None, "status": "needed"}


def shallow_merge_dicts(a, b):
    c = a.copy()
    c.update(b)
    return c


class TestCVE:
    def test_cve_init_attributes(self, FakeConfig):
        """CVE.__init__ saves client and response on instance."""
        client = UASecurityClient(FakeConfig())
        cve = CVE(client, {"some": "response"})
        assert client == cve.client
        assert {"some": "response"} == cve.response

    @pytest.mark.parametrize(
        "cve1,cve2,are_equal",
        (
            (CVE(None, {"1": "2"}), CVE(None, {"1": "2"}), True),
            (CVE("A", {"1": "2"}), CVE("B", {"1": "2"}), True),
            (CVE(None, {}), CVE("B", {"1": "2"}), False),
            (CVE(None, {"1": "2"}), USN(None, {"1": "2"}), False),
        ),
    )
    def test_equality(self, cve1, cve2, are_equal):
        """Equality is based instance type and CVE.response value"""
        if are_equal:
            assert cve1.response == cve2.response
            assert cve1 == cve2
        else:
            if isinstance(cve1, CVE) and isinstance(cve2, CVE):
                assert cve1.response != cve2.response
            assert cve1 != cve2

    @pytest.mark.parametrize(
        "attr_name,expected,response",
        (
            ("description", None, {}),
            ("description", "descr", {"description": "descr"}),
            ("id", "UNKNOWN_CVE_ID", {}),
            (
                "id",
                "CVE-123",
                {"id": "cve-123"},
            ),  # Uppercase of id value is used
            ("notices_ids", [], {}),
            ("notices_ids", [], {"notices_ids": []}),
            ("notices_ids", ["1", "2"], {"notices_ids": ["1", "2"]}),
        ),
    )
    def test_cve_basic_properties_from_response(
        self, attr_name, expected, response, FakeConfig
    ):
        """CVE instance properties are set from Security API CVE response."""
        client = UASecurityClient(FakeConfig())
        cve = CVE(client, response)
        assert expected == getattr(cve, attr_name)

    @pytest.mark.parametrize(
        "usns_response,expected",
        (
            (None, []),
            ([], []),
            (  # USNs are properly sorted by id
                [{"id": "USN-1"}, {"id": "USN-2"}, {"id": "LSN-3"}],
                [USN(None, {"id": "USN-2"}), USN(None, {"id": "USN-1"})],
            ),
        ),
    )
    def test_notices_cached_from_usns_response(
        self, usns_response, expected, FakeConfig
    ):
        """List of USNs returned from CVE 'usns' response if present."""
        client = UASecurityClient(FakeConfig())
        cve_response = copy.deepcopy(SAMPLE_CVE_RESPONSE)
        if usns_response is not None:
            cve_response["notices"] = usns_response
        cve = CVE(client, cve_response)
        assert expected == cve.notices
        # clear box test caching in effect
        cve.response = "junk"
        assert expected == cve.notices


class TestUSN:
    def test_usn_init_attributes(self, FakeConfig):
        """USN.__init__ saves client and response on instance."""
        client = UASecurityClient(FakeConfig())
        cve = USN(client, {"some": "response"})
        assert client == cve.client
        assert {"some": "response"} == cve.response

    @pytest.mark.parametrize(
        "usn1,usn2,are_equal",
        (
            (USN(None, {"1": "2"}), USN(None, {"1": "2"}), True),
            (USN("A", {"1": "2"}), USN("B", {"1": "2"}), True),
            (USN(None, {}), USN("B", {"1": "2"}), False),
            (USN(None, {"1": "2"}), CVE(None, {"1": "2"}), False),
        ),
    )
    def test_equality(self, usn1, usn2, are_equal):
        """Equality is based instance type and USN.response value"""
        if are_equal:
            assert usn1.response == usn2.response
            assert usn1 == usn2
        else:
            if isinstance(usn1, USN) and isinstance(usn2, USN):
                assert usn1.response != usn2.response
            assert usn1 != usn2

    @pytest.mark.parametrize(
        "attr_name,expected,response",
        (
            ("title", None, {}),
            ("title", "my title", {"title": "my title"}),
            ("id", "UNKNOWN_USN_ID", {}),
            (
                "id",
                "USN-123",
                {"id": "usn-123"},
            ),  # Uppercase of id value is used
            ("cves_ids", [], {}),
            ("cves_ids", [], {"cves_ids": []}),
            ("cves_ids", ["1", "2"], {"cves_ids": ["1", "2"]}),
            ("cves", [], {}),
            ("cves", [], {"cves": []}),
        ),
    )
    def test_usn_basic_properties_from_response(
        self, attr_name, expected, response, FakeConfig
    ):
        """USN instance properties are set from Security API USN response."""
        client = UASecurityClient(FakeConfig())
        usn = USN(client, response)
        assert expected == getattr(usn, attr_name)

    @pytest.mark.parametrize(
        "series,expected",
        (
            (
                "series-example-1",
                {
                    "samba": {
                        "source": {
                            "description": (
                                "SMB/CIFS file, print, and login ... Unix"
                            ),
                            "is_source": True,
                            "name": "samba",
                            "version": "2:4.3.11+dfsg-0ubuntu0.14.04.20+esm9",
                        },
                        "samba": {
                            "is_source": False,
                            "name": "samba",
                            "source_link": (
                                "https://launchpad.net/ubuntu/+source/samba"
                            ),
                            "version": "2~14.04.1+esm9",
                            "version_link": (
                                "https://....11+dfsg-0ubuntu0.14.04.20+esm9"
                            ),
                        },
                    }
                },
            ),
            (
                "series-example-2",
                {
                    "coin3": {
                        "source": {
                            "description": (
                                "high-level 3D graphics kit implementing ..."
                            ),
                            "is_source": True,
                            "name": "coin3",
                            "version": "3.1.4~abc9f50-4ubuntu2+esm1",
                        },
                        "libcoin80-runtime": {
                            "is_source": False,
                            "name": "libcoin80-runtime",
                            "source_link": (
                                "https://launchpad.net/ubuntu/+source/coin3"
                            ),
                            "version": "3~18.04.1+esm2",
                            "version_link": "https://coin3...18.04.1+esm2",
                        },
                    }
                },
            ),
            ("series-example-3", {}),
        ),
    )
    @mock.patch("uaclient.system.get_release_info")
    def test_release_packages_returns_source_and_binary_pkgs_for_series(
        self, m_get_release_info, series, expected, FakeConfig
    ):
        m_get_release_info.return_value = mock.MagicMock(series=series)
        client = UASecurityClient(FakeConfig())
        usn = USN(client, SAMPLE_USN_RESPONSE)

        assert expected == usn.release_packages
        usn._release_packages = {"sl": "1.0"}
        assert {"sl": "1.0"} == usn.release_packages

    @pytest.mark.parametrize(
        "source_link,error_msg",
        (
            (
                None,
                (
                    "Metadata for USN-4510-2 is invalid. "
                    "Error: USN-4510-2 metadata does not define "
                    "release_packages source_link for samba2."
                ),
            ),
            (
                "unknown format",
                (
                    "Metadata for USN-4510-2 is invalid. "
                    "Error: USN-4510-2 metadata has unexpected "
                    "release_packages source_link value for samba2: "
                    "unknown format."
                ),
            ),
        ),
    )
    @mock.patch("uaclient.system.get_release_info")
    def test_release_packages_errors_on_sparse_source_url(
        self, m_get_release_info, source_link, error_msg, FakeConfig
    ):
        """Raise errors when USN metadata contains no valid source_link."""
        m_get_release_info.return_value = mock.MagicMock(
            series="series-example-1"
        )
        client = UASecurityClient(FakeConfig())
        sparse_md = copy.deepcopy(SAMPLE_USN_RESPONSE)
        sparse_md["release_packages"]["series-example-1"].append(
            {
                "is_source": False,
                "name": "samba2",
                "source_link": source_link,
                "version": "2~14.04.1+esm9",
                "version_link": "https://....11+dfsg-0ubuntu0.14.04.20+esm9",
            }
        )
        usn = USN(client, sparse_md)
        with pytest.raises(exceptions.SecurityAPIMetadataError) as exc:
            usn.release_packages
        assert error_msg in str(exc.value)

    @pytest.mark.parametrize(
        "cves_response,expected",
        (
            (None, []),
            ([], []),
            (  # CVEs are properly sorted by id
                [{"id": "1"}, {"id": "2"}],
                [CVE(None, {"id": "2"}), CVE(None, {"id": "1"})],
            ),
        ),
    )
    def test_cves_cached_and_sorted_from_cves_response(
        self, cves_response, expected, FakeConfig
    ):
        """List of USNs returned from CVE 'usns' response if present."""
        client = UASecurityClient(FakeConfig())
        usn_response = copy.deepcopy(SAMPLE_USN_RESPONSE)
        if cves_response is not None:
            usn_response["cves"] = cves_response
        usn = USN(client, usn_response)
        assert expected == usn.cves
        # clear box test caching in effect
        usn.response = "junk"
        assert expected == usn.cves


class TestCVEPackageStatus:
    def test_simple_properties_from_response(self):
        pkg_status = CVEPackageStatus(
            cve_response=CVE_ESM_PACKAGE_STATUS_RESPONSE
        )
        assert CVE_ESM_PACKAGE_STATUS_RESPONSE == pkg_status.response
        assert pkg_status.response["description"] == pkg_status.description
        assert pkg_status.description == pkg_status.fixed_version
        assert pkg_status.response["pocket"] == pkg_status.pocket
        assert (
            pkg_status.response["release_codename"]
            == pkg_status.release_codename
        )
        assert pkg_status.response["status"] == pkg_status.status

    @pytest.mark.parametrize(
        "pocket,description,expected",
        (
            ("esm-infra", "1.2", "Ubuntu Pro: ESM Infra"),
            ("esm-apps", "1.2", "Ubuntu Pro: ESM Apps"),
            ("updates", "1.2esm", "Ubuntu standard updates"),
            ("security", "1.2esm", "Ubuntu standard updates"),
            (None, "1.2", "Ubuntu standard updates"),
            (None, "1.2esm", "Ubuntu Pro: ESM Infra"),
        ),
    )
    def test_pocket_source_from_response(self, pocket, description, expected):
        cve_response = {"pocket": pocket, "description": description}
        pkg_status = CVEPackageStatus(cve_response=cve_response)
        assert expected == pkg_status.pocket_source

    @pytest.mark.parametrize(
        "pocket,description,expected",
        (
            ("esm-infra", "1.2", True),
            ("esm-apps", "1.2", True),
            ("updates", "1.2esm", False),
            ("security", "1.2esm", False),
            (None, "1.2", False),
            (None, "1.2esm", True),
        ),
    )
    def test_requires_pro_from_response(self, pocket, description, expected):
        """requires_pro is derived from response pocket and description."""
        cve_response = {"pocket": pocket, "description": description}
        pkg_status = CVEPackageStatus(cve_response=cve_response)
        assert expected is pkg_status.requires_ua

    @pytest.mark.parametrize(
        "status,pocket,expected",
        (
            (
                "not-affected",
                "",
                "Source package is not affected on this release.",
            ),
            ("DNE", "", "Source package does not exist on this release."),
            (
                "needs-triage",
                "esm-infra",
                "Ubuntu security engineers are investigating this issue.",
            ),
            ("needed", "esm-infra", "Sorry, no fix is available yet."),
            (
                "pending",
                "esm-infra",
                "A fix is coming soon. Try again tomorrow.",
            ),
            ("ignored", "esm-infra", "Sorry, no fix is available."),
            (
                "released",
                "esm-infra",
                "A fix is available in Ubuntu Pro: ESM Infra.",
            ),
            (
                "released",
                "security",
                "A fix is available in Ubuntu standard updates.",
            ),
            ("bogus", "1.2", "UNKNOWN: bogus"),
        ),
    )
    def test_status_message_from_response(self, status, pocket, expected):
        cve_response = {"pocket": pocket, "status": status}
        pkg_status = CVEPackageStatus(cve_response=cve_response)
        assert expected == pkg_status.status_message


@mock.patch(M_PATH + "UASecurityClient.request_url")
class TestUASecurityClient:
    @pytest.mark.parametrize(
        "m_kwargs,expected_error, extra_security_params",
        (
            ({}, None, None),
            ({"query": "vq"}, None, {"test": "blah"}),
            (SAMPLE_GET_CVES_QUERY_PARAMS, None, None),
            ({"invalidparam": "vv"}, TypeError, None),
        ),
    )
    def test_get_cves_sets_query_params_on_get_cves_route(
        self,
        request_url,
        m_kwargs,
        expected_error,
        extra_security_params,
        FakeConfig,
    ):
        """GET CVE instances from API_V1_CVES route with querystrings"""
        cfg = FakeConfig()
        if extra_security_params:
            cfg.override_features(
                {"extra_security_params": extra_security_params}
            )

        client = UASecurityClient(cfg)
        if expected_error:
            with pytest.raises(expected_error) as exc:
                client.get_cves(**m_kwargs)
            assert (
                "get_cves() got an unexpected keyword argument 'invalidparam'"
            ) in str(exc.value)
            assert 0 == request_url.call_count
        else:
            for key in SAMPLE_GET_CVES_QUERY_PARAMS:
                if key not in m_kwargs:
                    m_kwargs[key] = None
            request_url.return_value = http.HTTPResponse(
                code=200,
                headers={},
                body="",
                json_dict={},
                json_list=["body1", "body2"],
            )
            [cve1, cve2] = client.get_cves(**m_kwargs)
            assert isinstance(cve1, CVE)
            assert isinstance(cve2, CVE)
            assert "body1" == cve1.response
            assert "body2" == cve2.response
            # get_cves transposes "query" to "q"
            m_kwargs["q"] = m_kwargs.pop("query")

            assert [
                mock.call(API_V1_CVES, query_params=m_kwargs)
            ] == request_url.call_args_list

    @pytest.mark.parametrize(
        "m_kwargs,expected_error, extra_security_params",
        (
            ({}, None, None),
            ({"cves": "cve"}, None, None),
            (SAMPLE_GET_NOTICES_QUERY_PARAMS, None, {"test": "blah"}),
            ({"invalidparam": "vv"}, TypeError, None),
        ),
    )
    def test_get_notices_sets_query_params_on_get_cves_route(
        self,
        request_url,
        m_kwargs,
        expected_error,
        extra_security_params,
        FakeConfig,
    ):
        """GET body from API_V1_NOTICES route with appropriate querystring"""
        cfg = FakeConfig()
        if extra_security_params:
            cfg.override_features(
                {"extra_security_params": extra_security_params}
            )

        client = UASecurityClient(cfg)
        if expected_error:
            with pytest.raises(expected_error) as exc:
                client.get_notices(**m_kwargs)
            assert (
                "get_notices() got an unexpected keyword argument"
                " 'invalidparam'"
            ) in str(exc.value)
            assert 0 == request_url.call_count
        else:
            for key in SAMPLE_GET_NOTICES_QUERY_PARAMS:
                if key not in m_kwargs:
                    m_kwargs[key] = None
            request_url.return_value = http.HTTPResponse(
                code=200,
                headers={},
                body="",
                json_dict={
                    "notices": [
                        {"id": "USN-2", "cves_ids": ["cve"]},
                        {"id": "USN-1", "cves_ids": ["cve"]},
                        {"id": "LSN-3", "cves_ids": ["cve"]},
                    ]
                },
                json_list=[],
            )
            [usn1, usn2] = client.get_notices(**m_kwargs)
            assert isinstance(usn1, USN)
            assert isinstance(usn2, USN)
            assert "USN-1" == usn1.id
            assert "USN-2" == usn2.id
            assert [
                mock.call(API_V1_NOTICES, query_params=m_kwargs)
            ] == request_url.call_args_list

    @pytest.mark.parametrize("cves", (("cve1"), (None)))
    def test_get_notices_filter_usns_when_setting_cves_param(
        self, request_url, cves, FakeConfig
    ):
        """Test if cves are used to filter the returned USNs."""
        cfg = FakeConfig()
        client = UASecurityClient(cfg)
        request_url.return_value = http.HTTPResponse(
            code=200,
            headers={},
            body="",
            json_dict={
                "notices": [
                    {"id": "USN-2", "cves_ids": ["cve2"]},
                    {"id": "USN-1", "cves_ids": ["cve1"]},
                    {"id": "LSN-3", "cves_ids": ["cve3"]},
                ]
            },
            json_list=[],
        )
        usns = client.get_notices(cves=cves)

        if cves:
            assert len(usns) == 1
            assert usns[0].id == "USN-1"
        else:
            assert len(usns) == 2
            assert usns[0].id == "USN-1"
            assert usns[1].id == "USN-2"

    @pytest.mark.parametrize(
        "m_kwargs,expected_error, extra_security_params",
        (({}, TypeError, None), ({"cve_id": "CVE-1"}, None, {"test": "blah"})),
    )
    def test_get_cve_provides_response_from_cve_json_route(
        self,
        request_url,
        m_kwargs,
        expected_error,
        extra_security_params,
        FakeConfig,
    ):
        """GET body from API_V1_CVE_TMPL route with required cve_id."""
        cfg = FakeConfig()
        if extra_security_params:
            cfg.override_features(
                {"extra_security_params": extra_security_params}
            )
        client = UASecurityClient(cfg)
        if expected_error:
            with pytest.raises(expected_error) as exc:
                client.get_cve(**m_kwargs)
            assert (
                "get_cve() missing 1 required positional argument: 'cve_id'"
            ) in str(exc.value)
            assert 0 == request_url.call_count
        else:
            request_url.return_value = http.HTTPResponse(
                code=200,
                headers={},
                body="",
                json_dict={"body": "body"},
                json_list=[],
            )
            cve = client.get_cve(**m_kwargs)
            assert isinstance(cve, CVE)
            assert {"body": "body"} == cve.response
            assert [
                mock.call(API_V1_CVE_TMPL.format(cve=m_kwargs["cve_id"]))
            ] == request_url.call_args_list

    @pytest.mark.parametrize(
        "m_kwargs,expected_error, extra_security_params",
        (
            ({}, TypeError, None),
            ({"notice_id": "USN-1"}, None, {"test": "blah"}),
        ),
    )
    def test_get_notice_provides_response_from_notice_json_route(
        self,
        request_url,
        m_kwargs,
        expected_error,
        extra_security_params,
        FakeConfig,
    ):
        """GET body from API_V1_NOTICE_TMPL route with required notice_id."""
        cfg = FakeConfig()
        if extra_security_params:
            cfg.override_features(
                {"extra_security_params": extra_security_params}
            )

        client = UASecurityClient(cfg)
        if expected_error:
            with pytest.raises(expected_error) as exc:
                client.get_notice(**m_kwargs)
            assert (
                "get_notice() missing 1 required positional argument:"
                " 'notice_id'"
            ) in str(exc.value)
            assert 0 == request_url.call_count
        else:
            request_url.return_value = http.HTTPResponse(
                code=200,
                headers={},
                body="",
                json_dict={"body": "body"},
                json_list=[],
            )
            assert {"body": "body"} == client.get_notice(**m_kwargs).response
            assert [
                mock.call(
                    API_V1_NOTICE_TMPL.format(notice=m_kwargs["notice_id"])
                )
            ] == request_url.call_args_list


class TestGetCVEAffectedPackageStatus:
    @pytest.mark.parametrize(
        "series,installed_packages,expected_status",
        (
            ("bionic", {}, {}),
            # installed package version has no bearing on status filtering
            ("bionic", {"samba": "1000"}, SAMBA_CVE_STATUS_BIONIC),
            # active series has a bearing on status filtering
            ("upstream", {"samba": "1000"}, SAMBA_CVE_STATUS_UPSTREAM),
            # package status not-affected gets filtered from affected_pkgs
            ("focal", {"samba": "1000"}, {}),
        ),
    )
    @mock.patch("uaclient.system.get_release_info")
    def test_affected_packages_status_filters_by_installed_pkgs_and_series(
        self,
        m_get_release_info,
        series,
        installed_packages,
        expected_status,
        FakeConfig,
    ):
        """Package statuses are filtered if not installed"""
        m_get_release_info.return_value = mock.MagicMock(series=series)
        client = UASecurityClient(FakeConfig())
        cve = CVE(client, SAMPLE_CVE_RESPONSE)
        affected_packages = get_cve_affected_source_packages_status(
            cve, installed_packages=installed_packages
        )
        if expected_status:
            package_status = affected_packages["samba"]
            assert expected_status == package_status.response
        else:
            assert expected_status == affected_packages


class TestQueryInstalledPkgSources:
    @pytest.mark.parametrize(
        "dpkg_out,results",
        (
            # Ignore b non-installed status
            ("a,,1.2,installed\nb,b,1.2,config-files", {"a": {"a": "1.2"}}),
            # Handle cases where no Source is defined for the pkg
            (
                "a,,1.2,installed\nzip,zip,3.0,installed",
                {"a": {"a": "1.2"}, "zip": {"zip": "3.0"}},
            ),
            # Prefer Source package name to binary package name
            (
                "b,bsrc,1.2,installed\nzip,zip,3.0,installed",
                {"bsrc": {"b": "1.2"}, "zip": {"zip": "3.0"}},
            ),
        ),
    )
    @mock.patch(M_PATH + "system.subp")
    @mock.patch("uaclient.system.get_release_info")
    def test_result_keyed_by_source_package_name(
        self, m_get_release_info, subp, dpkg_out, results
    ):
        m_get_release_info.return_value = mock.MagicMock(series="bionic")
        subp.return_value = dpkg_out, ""
        assert results == query_installed_source_pkg_versions()
        _format = "-f=${Package},${Source},${Version},${db:Status-Status}\n"
        assert [
            mock.call(["dpkg-query", _format, "-W"])
        ] == subp.call_args_list


class TestGetRelatedUSNs:
    def test_no_usns_returned_when_no_cves_are_found(self, FakeConfig):
        cfg = FakeConfig()
        client = UASecurityClient(cfg=cfg)
        usn = USN(client, SAMPLE_USN_RESPONSE_NO_CVES)

        assert [] == get_related_usns(usn, client)

    def test_usns_ignore_non_usns_items(self, FakeConfig):
        expected_value = mock.MagicMock(id="USN-1235-1")

        def fake_get_notice(notice_id):
            return expected_value

        m_client = mock.MagicMock()
        m_client.get_notice.side_effect = fake_get_notice

        m_usn = mock.MagicMock(
            cves=[
                mock.MagicMock(
                    notices_ids=["USN-1235-1", "LSN-0088-1"],
                )
            ],
            id="USN-8796-1",
        )

        assert [expected_value] == get_related_usns(m_usn, m_client)


class TestGetUSNAffectedPackagesStatus:
    @pytest.mark.parametrize(
        "installed_packages, affected_packages",
        (
            (
                {"coin3": {"libcoin80-runtime", "1.0"}},
                {
                    "coin3": CVEPackageStatus(
                        defaultdict(
                            str, {"status": "released", "pocket": "security"}
                        )
                    )
                },
            ),
        ),
    )
    @mock.patch("uaclient.system.get_release_info")
    def test_pkgs_come_from_release_packages_if_usn_has_no_cves(
        self,
        m_get_release_info,
        installed_packages,
        affected_packages,
        FakeConfig,
    ):
        m_get_release_info.return_value = mock.MagicMock(series="bionic")

        cfg = FakeConfig()
        client = UASecurityClient(cfg=cfg)
        usn = USN(client, SAMPLE_USN_RESPONSE_NO_CVES)
        actual_value = get_usn_affected_packages_status(
            usn, installed_packages
        )

        if not affected_packages:
            assert actual_value is {}
        else:
            assert "coin3" in actual_value
            assert (
                affected_packages["coin3"].status
                == actual_value["coin3"].status
            )
            assert (
                affected_packages["coin3"].pocket_source
                == actual_value["coin3"].pocket_source
            )


class TestOverrideUSNReleasePackageStatus:
    @pytest.mark.parametrize(
        "pkg_status",
        (
            CVE_PKG_STATUS_IGNORED,
            CVE_PKG_STATUS_PENDING,
            CVE_PKG_STATUS_NEEDS_TRIAGE,
            CVE_PKG_STATUS_NEEDED,
            CVE_PKG_STATUS_DEFERRED,
            CVE_PKG_STATUS_RELEASED,
            CVE_PKG_STATUS_RELEASED_ESM_INFRA,
        ),
    )
    @pytest.mark.parametrize(
        "usn_src_released_pkgs,expected",
        (
            ({}, None),
            (  # No "source" key, so ignore all binaries
                {"somebinary": {"pocket": "my-pocket", "version": "usn-ver"}},
                None,
            ),
            (
                {
                    "source": {
                        "name": "srcpkg",
                        "version": "usn-source-pkg-ver",
                    },
                    "somebinary": {
                        "pocket": "my-pocket",
                        "version": "usn-bin-ver",
                    },
                },
                {
                    "pocket": "my-pocket",
                    "description": "usn-source-pkg-ver",
                    "status": "released",
                },
            ),
        ),
    )
    def test_override_cve_src_info_with_pocket_and_ver_from_usn(
        self, usn_src_released_pkgs, expected, pkg_status
    ):
        """Override CVEPackageStatus with released/pocket from USN."""
        orig_cve = CVEPackageStatus(pkg_status)
        override = override_usn_release_package_status(
            orig_cve, usn_src_released_pkgs
        )
        if expected is None:  # Expect CVEPackageStatus unaltered
            assert override.response == orig_cve.response
        else:
            assert expected == override.response


class TestMergeUSNReleasedBinaryPackageVersions:
    @pytest.mark.parametrize(
        "usns_released_packages, expected_pkgs_dict",
        (
            ([{}], {}),
            (
                [{"pkg1": {"libpkg1": {"version": "1.0", "name": "libpkg1"}}}],
                {"pkg1": {"libpkg1": {"version": "1.0", "name": "libpkg1"}}},
            ),
            (
                [
                    {
                        "pkg1": {
                            "libpkg1": {"version": "1.0", "name": "libpkg1"}
                        },
                        "pkg2": {
                            "libpkg2": {"version": "2.0", "name": "libpkg2"},
                            "libpkg3": {"version": "3.0", "name": "libpkg3"},
                            "libpkg4": {"version": "3.0", "name": "libpkg4"},
                        },
                    },
                    {
                        "pkg2": {
                            "libpkg2": {"version": "1.8", "name": "libpkg2"},
                            "libpkg4": {"version": "3.2", "name": "libpkg4"},
                        }
                    },
                ],
                {
                    "pkg1": {"libpkg1": {"version": "1.0", "name": "libpkg1"}},
                    "pkg2": {
                        "libpkg2": {"version": "2.0", "name": "libpkg2"},
                        "libpkg3": {"version": "3.0", "name": "libpkg3"},
                        "libpkg4": {"version": "3.2", "name": "libpkg4"},
                    },
                },
            ),
            (
                [
                    {
                        "pkg1": {
                            "libpkg1": {"version": "1.0", "name": "libpkg1"},
                            "source": {"version": "2.0", "name": "pkg1"},
                        }
                    },
                    {"pkg1": {"source": {"version": "2.5", "name": "pkg1"}}},
                ],
                {
                    "pkg1": {
                        "libpkg1": {"version": "1.0", "name": "libpkg1"},
                        "source": {"version": "2.5", "name": "pkg1"},
                    }
                },
            ),
            (
                [
                    {
                        "pkg1": {
                            "libpkg1": {"version": "1.0", "name": "libpkg1"},
                            "source": {"version": "2.0", "name": "pkg1"},
                        },
                        "pkg2": {
                            "libpkg2": {
                                "version": "2.0",
                                "name": "libpkg2",
                                "pocket": "esm-apps",
                            },
                            "source": {
                                "version": "2.0",
                                "name": "pkg2",
                                "pocket": "esm-apps",
                            },
                        },
                    }
                ],
                {
                    "pkg1": {
                        "libpkg1": {"version": "1.0", "name": "libpkg1"},
                        "source": {"version": "2.0", "name": "pkg1"},
                    }
                },
            ),
        ),
    )
    def test_merge_usn_released_binary_package_versions(
        self, usns_released_packages, expected_pkgs_dict, _subp
    ):
        usns = []
        beta_packages = {"esm-infra": False, "esm-apps": True}

        for usn_released_pkgs in usns_released_packages:
            usn = mock.MagicMock()
            type(usn).release_packages = mock.PropertyMock(
                return_value=usn_released_pkgs
            )
            usns.append(usn)

        with mock.patch("uaclient.system._subp", side_effect=_subp):
            usn_pkgs_dict = merge_usn_released_binary_package_versions(
                usns, beta_packages
            )
        assert expected_pkgs_dict == usn_pkgs_dict



import mock
import pytest

from uaclient import exceptions
from uaclient.data_types import (
    DataObject,
    Field,
    IncorrectFieldTypeError,
    IntDataValue,
    StringDataValue,
)
from uaclient.files.data_types import DataObjectFile, DataObjectFileFormat


class MockUAFile:
    def __init__(self):
        self.write = mock.MagicMock()
        self.read = mock.MagicMock()
        self.delete = mock.MagicMock()
        self.path = mock.MagicMock()


class NestedTestData(DataObject):
    fields = [
        Field("integer", IntDataValue),
    ]

    def __init__(self, integer: int):
        self.integer = integer


class TestData(DataObject):
    fields = [
        Field("string", StringDataValue),
        Field("nested", NestedTestData),
    ]

    def __init__(self, string: str, nested: NestedTestData):
        self.string = string
        self.nested = nested


class TestDataObjectFile:
    def test_write_valid_json(self):
        mock_file = MockUAFile()
        dof = DataObjectFile(
            TestData,
            mock_file,
        )
        dof.write(TestData(string="test", nested=NestedTestData(integer=1)))
        assert mock_file.write.call_args_list == [
            mock.call("""{"nested": {"integer": 1}, "string": "test"}""")
        ]

    def test_write_valid_yaml(self):
        mock_file = MockUAFile()
        dof = DataObjectFile(
            TestData,
            mock_file,
            DataObjectFileFormat.YAML,
        )
        dof.write(TestData(string="test", nested=NestedTestData(integer=1)))
        assert mock_file.write.call_args_list == [
            mock.call("""nested:\n  integer: 1\nstring: test\n""")
        ]

    def test_read_valid_json(self):
        mock_file = MockUAFile()
        dof = DataObjectFile(
            TestData,
            mock_file,
        )
        mock_file.read.return_value = (
            """{"string": "test", "nested": {"integer": 1}}"""
        )
        do = dof.read()
        assert do.string == "test"
        assert do.nested.integer == 1

    def test_read_valid_yaml(self):
        mock_file = MockUAFile()
        dof = DataObjectFile(
            TestData,
            mock_file,
            DataObjectFileFormat.YAML,
        )
        mock_file.read.return_value = (
            """nested:\n  integer: 1\nstring: test\n"""
        )
        do = dof.read()
        assert do.string == "test"
        assert do.nested.integer == 1

    def test_read_invalid_data(self):
        mock_file = MockUAFile()
        dof = DataObjectFile(
            TestData,
            mock_file,
        )
        mock_file.read.return_value = """{"nested": {"integer": 1}}"""
        with pytest.raises(IncorrectFieldTypeError):
            dof.read()

    def test_read_invalid_json(self):
        mock_file = MockUAFile()
        dof = DataObjectFile(
            TestData,
            mock_file,
        )
        mock_file.read.return_value = """{"nested": {"""
        with pytest.raises(exceptions.InvalidFileFormatError):
            dof.read()

    def test_read_invalid_yaml(self):
        mock_file = MockUAFile()
        dof = DataObjectFile(
            TestData,
            mock_file,
            DataObjectFileFormat.YAML,
        )
        mock_file.read.return_value = """nested": {"""
        with pytest.raises(exceptions.InvalidFileFormatError):
            dof.read()



import copy
import json

import mock
import pytest

from uaclient.entitlements import valid_services
from uaclient.files.machine_token import MachineTokenFile


@pytest.fixture
def all_resources_available(FakeConfig):
    resources = [
        {"name": name, "available": True}
        for name in valid_services(cfg=FakeConfig())
    ]
    return resources


@pytest.fixture
def all_resources_entitled(FakeConfig):
    resources = [
        {"type": name, "entitled": True}
        for name in valid_services(cfg=FakeConfig())
    ]
    return resources


@pytest.fixture
def no_resources_entitled(FakeConfig):
    resources = [
        {"type": name, "entitled": False}
        for name in valid_services(cfg=FakeConfig())
    ]
    return resources


@pytest.fixture
def resp_only_fips_resource_available(FakeConfig):
    resources = [
        {"name": name, "available": name == "fips"}
        for name in valid_services(cfg=FakeConfig())
    ]
    return resources


class TestAccounts:
    @mock.patch("uaclient.files.machine_token.MachineTokenFile.read")
    def test_accounts_returns_none_when_no_cached_account_value(
        self, m_machine_token_read, all_resources_available
    ):
        """Config.accounts property returns an empty list when no cache."""
        machine_token_file = MachineTokenFile()
        m_machine_token_read.return_value = None
        assert machine_token_file.account == {}

    @pytest.mark.usefixtures("all_resources_available")
    def test_accounts_extracts_account_key_from_machine_token_cache(
        self, all_resources_available, tmpdir, FakeConfig
    ):
        """Use machine_token cached accountInfo when no accounts cache."""
        machine_token_file = MachineTokenFile()
        accountInfo = {"id": "1", "name": "accountname"}

        machine_token_file._machine_token = {
            "availableResources": all_resources_available,
            "machineTokenInfo": {"accountInfo": accountInfo},
        }

        assert accountInfo == machine_token_file.account


class TestMachineTokenOverlay:
    machine_token_dict = {
        "availableResources": [
            {"available": False, "name": "cc-eal"},
            {"available": True, "name": "esm-infra"},
            {"available": False, "name": "fips"},
        ],
        "machineTokenInfo": {
            "contractInfo": {
                "resourceEntitlements": [
                    {
                        "type": "cc-eal",
                        "entitled": False,
                        "affordances": {
                            "architectures": [
                                "amd64",
                                "ppc64el",
                                "ppc64le",
                                "s390x",
                                "x86_64",
                            ],
                            "series": ["xenial"],
                        },
                        "directives": {
                            "additionalPackages": ["ubuntu-commoncriteria"],
                            "aptKey": "key",
                            "aptURL": "https://esm.ubuntu.com/cc",
                            "suites": ["xenial"],
                        },
                    },
                    {
                        "type": "livepatch",
                        "entitled": True,
                        "affordances": {
                            "architectures": ["amd64", "x86_64"],
                            "tier": "stable",
                        },
                        "directives": {
                            "caCerts": "",
                            "remoteServer": "https://livepatch.canonical.com",
                        },
                        "obligations": {"enableByDefault": True},
                    },
                ]
            }
        },
    }

    @mock.patch("uaclient.system.load_file")
    @mock.patch("uaclient.files.machine_token.MachineTokenFile.read")
    @mock.patch("uaclient.config.os.path.exists", return_value=True)
    def test_machine_token_update_with_overlay(
        self, m_path, m_token_read, m_load_file, FakeConfig
    ):
        m_token_read.return_value = self.machine_token_dict

        remote_server_overlay = "overlay"
        json_str = json.dumps(
            {
                "availableResources": [
                    {"available": False, "name": "esm-infra"},
                    {"available": True, "name": "test-overlay"},
                ],
                "machineTokenInfo": {
                    "contractInfo": {
                        "resourceEntitlements": [
                            {
                                "type": "livepatch",
                                "entitled": False,
                                "affordances": {"architectures": ["test"]},
                                "directives": {"remoteServer": "overlay"},
                            }
                        ]
                    }
                },
            }
        )
        m_load_file.return_value = json_str

        expected = copy.deepcopy(self.machine_token_dict)
        expected["machineTokenInfo"]["contractInfo"]["resourceEntitlements"][
            1
        ]["directives"]["remoteServer"] = remote_server_overlay
        expected["machineTokenInfo"]["contractInfo"]["resourceEntitlements"][
            1
        ]["affordances"]["architectures"] = ["test"]
        expected["machineTokenInfo"]["contractInfo"]["resourceEntitlements"][
            1
        ]["entitled"] = False
        expected["availableResources"][1]["available"] = False
        expected["availableResources"].append(
            {"available": True, "name": "test-overlay"}
        )

        machine_token_file = MachineTokenFile(
            machine_token_overlay_path="test"
        )
        assert expected == machine_token_file.machine_token

    @mock.patch("uaclient.files.machine_token.MachineTokenFile.read")
    def test_machine_token_without_overlay(self, m_token_read, FakeConfig):
        m_token_read.return_value = self.machine_token_dict
        machine_token_file = MachineTokenFile()
        assert self.machine_token_dict == machine_token_file.machine_token


class TestEntitlements:
    def test_entitlements_property_keyed_by_entitlement_name(
        self, all_resources_available
    ):
        """Return machine_token resourceEntitlements, keyed by name."""
        machine_token_file = MachineTokenFile()
        machine_token_file._machine_token = {
            "availableResources": all_resources_available,
            "machineTokenInfo": {
                "contractInfo": {
                    "resourceEntitlements": [
                        {"type": "entitlement1", "entitled": True},
                        {"type": "entitlement2", "entitled": True},
                    ]
                }
            },
        }
        expected = {
            "entitlement1": {
                "entitlement": {"entitled": True, "type": "entitlement1"}
            },
            "entitlement2": {
                "entitlement": {"entitled": True, "type": "entitlement2"}
            },
        }
        assert expected == machine_token_file.entitlements()

    def test_entitlements_uses_resource_token_from_machine_token(
        self, all_resources_available
    ):
        """Include entitlement-specific resourceTokens from machine_token"""
        machine_token_file = MachineTokenFile()
        machine_token_file._machine_token = {
            "availableResources": all_resources_available,
            "machineTokenInfo": {
                "contractInfo": {
                    "resourceEntitlements": [
                        {"type": "entitlement1", "entitled": True},
                        {"type": "entitlement2", "entitled": True},
                    ]
                }
            },
            "resourceTokens": [
                {"type": "entitlement1", "token": "ent1-token"},
                {"type": "entitlement2", "token": "ent2-token"},
            ],
        }
        expected = {
            "entitlement1": {
                "entitlement": {"entitled": True, "type": "entitlement1"},
                "resourceToken": "ent1-token",
            },
            "entitlement2": {
                "entitlement": {"entitled": True, "type": "entitlement2"},
                "resourceToken": "ent2-token",
            },
        }
        assert expected == machine_token_file.entitlements()



import mock

from uaclient import messages
from uaclient.api.u.pro.services.dependencies.v1 import (
    DependenciesResult,
    Reason,
    ServiceWithDependencies,
    ServiceWithReason,
    _dependencies,
)


class TestServicesDependenciesV1:
    def test_dependencies(self):
        assert _dependencies(cfg=mock.MagicMock()) == DependenciesResult(
            services=[
                ServiceWithDependencies(
                    name="anbox-cloud", incompatible_with=[], depends_on=[]
                ),
                ServiceWithDependencies(
                    name="cc-eal", incompatible_with=[], depends_on=[]
                ),
                ServiceWithDependencies(
                    name="cis", incompatible_with=[], depends_on=[]
                ),
                ServiceWithDependencies(
                    name="esm-apps", incompatible_with=[], depends_on=[]
                ),
                ServiceWithDependencies(
                    name="esm-apps-legacy",
                    incompatible_with=[],
                    depends_on=[],
                ),
                ServiceWithDependencies(
                    name="esm-infra", incompatible_with=[], depends_on=[]
                ),
                ServiceWithDependencies(
                    name="esm-infra-legacy",
                    incompatible_with=[],
                    depends_on=[],
                ),
                ServiceWithDependencies(
                    name="fips",
                    incompatible_with=[
                        ServiceWithReason(
                            name="livepatch",
                            reason=Reason(
                                code=messages.LIVEPATCH_INVALIDATES_FIPS.name,
                                title=messages.LIVEPATCH_INVALIDATES_FIPS.msg,
                            ),
                        ),
                        ServiceWithReason(
                            name="fips-updates",
                            reason=Reason(
                                code=messages.FIPS_UPDATES_INVALIDATES_FIPS.name,  # noqa: E501
                                title=messages.FIPS_UPDATES_INVALIDATES_FIPS.msg,  # noqa: E501
                            ),
                        ),
                        ServiceWithReason(
                            name="realtime-kernel",
                            reason=Reason(
                                code=messages.REALTIME_FIPS_INCOMPATIBLE.name,
                                title=messages.REALTIME_FIPS_INCOMPATIBLE.msg,
                            ),
                        ),
                    ],
                    depends_on=[],
                ),
                ServiceWithDependencies(
                    name="fips-updates",
                    incompatible_with=[
                        ServiceWithReason(
                            name="fips",
                            reason=Reason(
                                code=messages.FIPS_INVALIDATES_FIPS_UPDATES.name,  # noqa: E501
                                title=messages.FIPS_INVALIDATES_FIPS_UPDATES.msg,  # noqa: E501
                            ),
                        ),
                        ServiceWithReason(
                            name="realtime-kernel",
                            reason=Reason(
                                code=messages.REALTIME_FIPS_UPDATES_INCOMPATIBLE.name,  # noqa: E501
                                title=messages.REALTIME_FIPS_UPDATES_INCOMPATIBLE.msg,  # noqa: E501
                            ),
                        ),
                    ],
                    depends_on=[],
                ),
                ServiceWithDependencies(
                    name="fips-preview",
                    incompatible_with=[
                        ServiceWithReason(
                            name="livepatch",
                            reason=Reason(
                                code=messages.LIVEPATCH_INVALIDATES_FIPS.name,
                                title=messages.LIVEPATCH_INVALIDATES_FIPS.msg,
                            ),
                        ),
                        ServiceWithReason(
                            name="fips-updates",
                            reason=Reason(
                                code=messages.FIPS_UPDATES_INVALIDATES_FIPS.name,  # noqa: E501
                                title=messages.FIPS_UPDATES_INVALIDATES_FIPS.msg,  # noqa: E501
                            ),
                        ),
                        ServiceWithReason(
                            name="realtime-kernel",
                            reason=Reason(
                                code=messages.REALTIME_FIPS_INCOMPATIBLE.name,
                                title=messages.REALTIME_FIPS_INCOMPATIBLE.msg,
                            ),
                        ),
                        ServiceWithReason(
                            name="fips",
                            reason=Reason(
                                code=messages.FIPS_INVALIDATES_FIPS_UPDATES.name,  # noqa: E501
                                title=messages.FIPS_INVALIDATES_FIPS_UPDATES.msg,  # noqa: E501
                            ),
                        ),
                    ],
                    depends_on=[],
                ),
                ServiceWithDependencies(
                    name="landscape", incompatible_with=[], depends_on=[]
                ),
                ServiceWithDependencies(
                    name="livepatch",
                    incompatible_with=[
                        ServiceWithReason(
                            name="fips",
                            reason=Reason(
                                code=messages.LIVEPATCH_INVALIDATES_FIPS.name,
                                title=messages.LIVEPATCH_INVALIDATES_FIPS.msg,
                            ),
                        ),
                        ServiceWithReason(
                            name="realtime-kernel",
                            reason=Reason(
                                code=messages.REALTIME_LIVEPATCH_INCOMPATIBLE.name,  # noqa: E501
                                title=messages.REALTIME_LIVEPATCH_INCOMPATIBLE.msg,  # noqa: E501
                            ),
                        ),
                    ],
                    depends_on=[],
                ),
                ServiceWithDependencies(
                    name="realtime-kernel",
                    incompatible_with=[
                        ServiceWithReason(
                            name="fips",
                            reason=Reason(
                                code=messages.REALTIME_FIPS_INCOMPATIBLE.name,
                                title=messages.REALTIME_FIPS_INCOMPATIBLE.msg,
                            ),
                        ),
                        ServiceWithReason(
                            name="fips-updates",
                            reason=Reason(
                                code=messages.REALTIME_FIPS_UPDATES_INCOMPATIBLE.name,  # noqa: E501
                                title=messages.REALTIME_FIPS_UPDATES_INCOMPATIBLE.msg,  # noqa: E501
                            ),
                        ),
                        ServiceWithReason(
                            name="livepatch",
                            reason=Reason(
                                code=messages.REALTIME_LIVEPATCH_INCOMPATIBLE.name,  # noqa: E501
                                title=messages.REALTIME_LIVEPATCH_INCOMPATIBLE.msg,  # noqa: E501
                            ),
                        ),
                    ],
                    depends_on=[],
                ),
                ServiceWithDependencies(
                    name="ros",
                    incompatible_with=[],
                    depends_on=[
                        ServiceWithReason(
                            name="esm-infra",
                            reason=Reason(
                                code=messages.ROS_REQUIRES_ESM.name,
                                title=messages.ROS_REQUIRES_ESM.msg,
                            ),
                        ),
                        ServiceWithReason(
                            name="esm-apps",
                            reason=Reason(
                                code=messages.ROS_REQUIRES_ESM.name,
                                title=messages.ROS_REQUIRES_ESM.msg,
                            ),
                        ),
                    ],
                ),
                ServiceWithDependencies(
                    name="ros-updates",
                    incompatible_with=[],
                    depends_on=[
                        ServiceWithReason(
                            name="esm-infra",
                            reason=Reason(
                                code=messages.ROS_REQUIRES_ESM.name,
                                title=messages.ROS_REQUIRES_ESM.msg,
                            ),
                        ),
                        ServiceWithReason(
                            name="esm-apps",
                            reason=Reason(
                                code=messages.ROS_REQUIRES_ESM.name,
                                title=messages.ROS_REQUIRES_ESM.msg,
                            ),
                        ),
                        ServiceWithReason(
                            name="ros",
                            reason=Reason(
                                code=messages.ROS_UPDATES_REQUIRES_ROS.name,
                                title=messages.ROS_UPDATES_REQUIRES_ROS.msg,
                            ),
                        ),
                    ],
                ),
            ]
        )



import logging
import os

import pycloudlib
from pycloudlib.cloud import ImageType


def handle_ssh_key(ec2, key_name):
    """Manage ssh keys to be used in the instances."""
    if key_name in ec2.list_keys():
        ec2.delete_key(key_name)

    key_pair = ec2.client.create_key_pair(KeyName=key_name)
    private_key_path = "ec2-test.pem"
    with open(private_key_path, "w", encoding="utf-8") as stream:
        stream.write(key_pair["KeyMaterial"])
    os.chmod(private_key_path, 0o600)

    # Since we are using a pem file, we don't have distinct public and
    # private key paths
    ec2.use_key(
        public_key_path=private_key_path,
        private_key_path=private_key_path,
        name=key_name,
    )


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    ec2 = pycloudlib.EC2(tag="examples")
    key_name = "test-ec2"
    handle_ssh_key(ec2, key_name)

    daily_pro = ec2.daily_image(release="focal", image_type=ImageType.PRO)

    print("Launching Pro instance...")
    instance = ec2.launch(daily_pro, instance_type="m5.large")
    instance.execute("touch custom_config_file")
    print(instance.execute("sudo ua status --wait"))
    print(instance.execute("sudo apt update"))
    print(instance.execute("sudo apt install ubuntu-advantage-tools"))

    print("")
    print("install ua version with the fix (must be at ./ua.deb)")
    print("")
    instance.push_file("./ua.deb", "/tmp/ua.deb")
    print(instance.execute("sudo dpkg -i /tmp/ua.deb"))

    print("")
    print("snapshotting")
    print("")
    image = ec2.snapshot(instance)

    print("")
    print("launching clone  - if this finishes, then success!")
    print("")
    new_instance = ec2.launch(image, instance_type="m5.large")
    print(new_instance.execute("sudo ua status --wait"))

    print("")
    print("Deleting Pro instances and image")
    print("")
    new_instance.delete()
    ec2.delete_image(image)
    instance.delete()



import mock
import pytest

from uaclient import event_logger, messages
from uaclient.api import exceptions
from uaclient.api.u.pro.attach.auto.full_auto_attach.v1 import (
    FullAutoAttachOptions,
    FullAutoAttachResult,
    _enable_services_by_name,
    _full_auto_attach,
    _full_auto_attach_in_lock,
)
from uaclient.entitlements.entitlement_status import (
    CanEnableFailure,
    CanEnableFailureReason,
)
from uaclient.testing import fakes
from uaclient.testing.helpers import does_not_raise

M_PATH = "uaclient.api.u.pro.attach.auto.full_auto_attach.v1."


class TestEnableServicesByName:
    @pytest.mark.parametrize(
        [
            "services",
            "enable_side_effect",
            "expected_enable_call_args",
            "expected_ret",
        ],
        [
            # success one service
            (
                ["esm-infra"],
                [(True, None)],
                [mock.call(mock.ANY, "esm-infra")],
                [],
            ),
            # success multi service
            (
                ["esm-apps", "esm-infra", "livepatch"],
                [(True, None), (True, None), (True, None)],
                [
                    mock.call(mock.ANY, "esm-apps"),
                    mock.call(
                        mock.ANY,
                        "esm-infra",
                    ),
                    mock.call(
                        mock.ANY,
                        "livepatch",
                    ),
                ],
                [],
            ),
            # fail via user facing error
            (
                ["esm-apps", "esm-infra", "livepatch"],
                [
                    (True, None),
                    exceptions.EntitlementNotFoundError(
                        entitlement_name="name"
                    ),
                    fakes.FakeUbuntuProError(),
                ],
                [
                    mock.call(mock.ANY, "esm-apps"),
                    mock.call(
                        mock.ANY,
                        "esm-infra",
                    ),
                    mock.call(
                        mock.ANY,
                        "livepatch",
                    ),
                ],
                [
                    (
                        "esm-infra",
                        messages.E_ENTITLEMENT_NOT_FOUND.format(
                            entitlement_name="name"
                        ),
                    ),
                    ("livepatch", fakes.FakeUbuntuProError._msg),
                ],
            ),
            # fail via return
            (
                ["esm-apps", "esm-infra", "livepatch"],
                [
                    (True, None),
                    (False, None),
                    (
                        False,
                        CanEnableFailure(
                            CanEnableFailureReason.ALREADY_ENABLED,
                            messages.NamedMessage("test", "test"),
                        ),
                    ),
                ],
                [
                    mock.call(mock.ANY, "esm-apps"),
                    mock.call(
                        mock.ANY,
                        "esm-infra",
                    ),
                    mock.call(
                        mock.ANY,
                        "livepatch",
                    ),
                ],
                [
                    (
                        "esm-infra",
                        messages.NamedMessage("unknown", "failed to enable"),
                    ),
                    ("livepatch", messages.NamedMessage("test", "test")),
                ],
            ),
        ],
    )
    @mock.patch(M_PATH + "actions.enable_entitlement_by_name")
    def test_enable_services_by_name(
        self,
        m_enable_entitlement_by_name,
        services,
        enable_side_effect,
        expected_enable_call_args,
        expected_ret,
    ):
        m_enable_entitlement_by_name.side_effect = enable_side_effect
        ret = _enable_services_by_name(mock.MagicMock(), services)
        assert (
            m_enable_entitlement_by_name.call_args_list
            == expected_enable_call_args
        )
        assert ret == expected_ret


@mock.patch("uaclient.files.notices.add")
@mock.patch("uaclient.files.notices.remove")
class TestFullAutoAttachV1:
    @mock.patch("uaclient.lock.RetryLock.__enter__")
    @mock.patch(
        M_PATH + "contract.UAContractClient.update_activity_token",
    )
    @mock.patch(
        "uaclient.actions.enable_entitlement_by_name",
    )
    @mock.patch("uaclient.clouds.identity.cloud_instance_factory")
    @mock.patch("uaclient.actions.auto_attach")
    def test_error_invalid_ent_names(
        self,
        _auto_attach,
        _cloud_instance_factory,
        m_enable_ent_by_name,
        _m_update_activity_token,
        _m_lock_enter,
        _notice_remove,
        _notice_add,
        FakeConfig,
    ):
        cfg = FakeConfig()

        def enable_ent_side_effect(cfg, name):
            if name != "wrong":
                return (True, None)

            return (False, None)

        m_enable_ent_by_name.side_effect = enable_ent_side_effect
        options = FullAutoAttachOptions(
            enable=["esm-infra", "cis"],
            enable_beta=["esm-apps", "realtime-kernel", "wrong"],
        )
        with pytest.raises(exceptions.EntitlementsNotEnabledError):
            _full_auto_attach(options, cfg)

        assert 5 == m_enable_ent_by_name.call_count

    @mock.patch("uaclient.lock.RetryLock.__enter__")
    @mock.patch(
        M_PATH + "contract.UAContractClient.update_activity_token",
    )
    @mock.patch(
        "uaclient.actions.enable_entitlement_by_name",
        return_value=(False, None),
    )
    @mock.patch("uaclient.clouds.identity.cloud_instance_factory")
    @mock.patch("uaclient.actions.auto_attach")
    def test_error_full_auto_attach_fail(
        self,
        _auto_attach,
        _cloud_instance_factory,
        enable_ent_by_name,
        _m_update_activity_token,
        _m_lock_enter,
        _notice_remove,
        _notice_add,
        FakeConfig,
    ):
        cfg = FakeConfig()
        options = FullAutoAttachOptions(
            enable=["esm-infra", "fips"],
            enable_beta=["esm-apps", "ros"],
        )
        with pytest.raises(exceptions.EntitlementsNotEnabledError):
            _full_auto_attach(options, cfg)

        assert 4 == enable_ent_by_name.call_count

    @mock.patch(
        "uaclient.lock.RetryLock.__enter__",
        side_effect=[
            exceptions.LockHeldError(
                lock_request="request", lock_holder="holder", pid=10
            ),
        ],
    )
    def test_lock_held(
        self, _m_retrylock_enter, _notice_remove, _notice_read, FakeConfig
    ):
        with pytest.raises(exceptions.LockHeldError):
            _full_auto_attach(FullAutoAttachOptions, FakeConfig())

    @pytest.mark.parametrize(
        "mode",
        list(map(lambda e: e.value, event_logger.EventLoggerMode)),
    )
    @pytest.mark.parametrize(
        [
            "options",
            "is_attached",
            "is_disabled",
            "expected_auto_attach_call_args",
            "enable_services_by_name_side_effect",
            "expected_enable_services_by_name_call_args",
            "raise_expectation",
            "expect_activity_call",
            "expected_error_message",
            "expected_ret",
        ],
        [
            # already attached
            (
                FullAutoAttachOptions(),
                True,
                False,
                [],
                [],
                [],
                pytest.raises(exceptions.AlreadyAttachedError),
                False,
                messages.E_ALREADY_ATTACHED.format(account_name="test").msg,
                None,
            ),
            # disable_auto_attach: true
            (
                FullAutoAttachOptions(),
                False,
                True,
                [],
                [],
                [],
                pytest.raises(exceptions.AutoAttachDisabledError),
                False,
                messages.E_AUTO_ATTACH_DISABLED_ERROR.msg,
                None,
            ),
            # success no options
            (
                FullAutoAttachOptions(),
                False,
                False,
                [mock.call(mock.ANY, mock.ANY, allow_enable=True)],
                [],
                [],
                does_not_raise(),
                True,
                None,
                FullAutoAttachResult(),
            ),
            # success enable
            (
                FullAutoAttachOptions(enable=["cis"]),
                False,
                False,
                [mock.call(mock.ANY, mock.ANY, allow_enable=False)],
                [[]],
                [mock.call(mock.ANY, ["cis"])],
                does_not_raise(),
                True,
                None,
                FullAutoAttachResult(),
            ),
            # success enable_beta
            (
                FullAutoAttachOptions(enable_beta=["cis"]),
                False,
                False,
                [mock.call(mock.ANY, mock.ANY, allow_enable=False)],
                [[]],
                [mock.call(mock.ANY, ["cis"])],
                does_not_raise(),
                True,
                None,
                FullAutoAttachResult(),
            ),
            # success enable and enable_beta
            (
                FullAutoAttachOptions(enable=["fips"], enable_beta=["cis"]),
                False,
                False,
                [mock.call(mock.ANY, mock.ANY, allow_enable=False)],
                [[], []],
                [
                    mock.call(mock.ANY, ["fips"]),
                    mock.call(mock.ANY, ["cis"]),
                ],
                does_not_raise(),
                True,
                None,
                FullAutoAttachResult(),
            ),
            # fail to enable
            (
                FullAutoAttachOptions(enable=["fips"], enable_beta=["cis"]),
                False,
                False,
                [mock.call(mock.ANY, mock.ANY, allow_enable=False)],
                [
                    [("fips", messages.NamedMessage("one", "two"))],
                    [("cis", messages.NamedMessage("three", "four"))],
                ],
                [
                    mock.call(mock.ANY, ["fips"]),
                    mock.call(mock.ANY, ["cis"]),
                ],
                pytest.raises(exceptions.EntitlementsNotEnabledError),
                True,
                messages.E_ENTITLEMENTS_NOT_ENABLED_ERROR.msg,
                None,
            ),
        ],
    )
    @mock.patch(
        M_PATH + "contract.UAContractClient.update_activity_token",
    )
    @mock.patch(M_PATH + "_enable_services_by_name")
    @mock.patch(M_PATH + "actions.auto_attach")
    @mock.patch(M_PATH + "identity.cloud_instance_factory")
    @mock.patch(M_PATH + "util.is_config_value_true")
    def test_full_auto_attach_v1(
        self,
        m_is_config_value_true,
        m_cloud_instance_factory,
        m_auto_attach,
        m_enable_services_by_name,
        m_update_activity_token,
        _notice_remove,
        _notice_add,
        options,
        is_attached,
        is_disabled,
        expected_auto_attach_call_args,
        enable_services_by_name_side_effect,
        expected_enable_services_by_name_call_args,
        raise_expectation,
        expect_activity_call,
        expected_error_message,
        expected_ret,
        mode,
        FakeConfig,
        fake_machine_token_file,
    ):
        fake_machine_token_file.attached = is_attached
        cfg = FakeConfig()
        m_is_config_value_true.return_value = is_disabled
        m_enable_services_by_name.side_effect = (
            enable_services_by_name_side_effect
        )
        with raise_expectation as e:
            ret = _full_auto_attach_in_lock(options, cfg, mode)
        assert m_auto_attach.call_args_list == expected_auto_attach_call_args
        assert (
            m_enable_services_by_name.call_args_list
            == expected_enable_services_by_name_call_args
        )
        if expected_error_message is not None:
            assert e.value.msg == expected_error_message
        if expected_ret is not None:
            assert ret == expected_ret
        if expect_activity_call:
            assert 1 == m_update_activity_token.call_count
        else:
            assert 0 == m_update_activity_token.call_count



import logging

import mock
import pytest

from uaclient.upgrade_lts_contract import process_contract_delta_after_apt_lock


@pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
class TestUpgradeLTSContract:
    @mock.patch("uaclient.upgrade_lts_contract._is_attached")
    def test_unattached_noops(
        self, m_is_attached, capsys, caplog_text, FakeConfig
    ):
        m_is_attached.return_value = mock.MagicMock(is_attached=False)
        expected_logs = [
            "Check whether to upgrade-lts-contract",
            "Skipping upgrade-lts-contract. Machine is unattached",
        ]

        process_contract_delta_after_apt_lock(FakeConfig())

        assert 1 == m_is_attached.call_count
        out, _err = capsys.readouterr()
        assert "" == out
        debug_logs = caplog_text()
        for log in expected_logs:
            assert log in debug_logs



#!/usr/bin/env python3

"""
This function is called from lib/upgrade_lts_contract.py and from
lib/reboot_cmds.py

This function should be used after running do-release-upgrade in a machine.
It will detect any contract deltas between the release before
do-release-upgrade and the current release. If we find any differences in
the uaclient contract between those releases, we will apply that difference
in the upgraded release.

For example, suppose we are on Trusty and we are upgrading to Xenial. We found
that the apt url for esm services on trusty:

https://esm.ubuntu.com/ubuntu

While on Xenial, the apt url is:

https://esm.ubuntu.com/infra/ubuntu

This script will detect differences like that and update the Xenial system
to reflect them.
"""

import fcntl
import logging

from uaclient import defaults, exceptions, messages, system, util
from uaclient.api import ProgressWrapper
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UAConfig
from uaclient.entitlements import (
    entitlement_factory,
    entitlements_enable_order,
)
from uaclient.entitlements.entitlement_status import (
    ApplicabilityStatus,
    ApplicationStatus,
)

# We consider the past release for LTSs to be the last LTS,
# because we don't have any services available on non-LTS.
# This makes it safer for us to try to process contract deltas.
# For example, we had "jammy": "focal" even when Impish was
# still supported.
current_codename_to_past_codename = {
    "xenial": "trusty",
    "bionic": "xenial",
    "focal": "bionic",
    "jammy": "focal",
    "noble": "jammy",
    "plucky": "noble",
    "questing": "plucky",
}

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


def process_contract_delta_after_apt_lock(cfg: UAConfig) -> None:
    LOG.debug("Check whether to upgrade-lts-contract")
    if not _is_attached(cfg).is_attached:
        LOG.debug("Skipping upgrade-lts-contract. Machine is unattached")
        return
    LOG.debug("Starting upgrade-lts-contract.")

    print(messages.RELEASE_UPGRADE_STARTING)

    print(messages.RELEASE_UPGRADE_APT_LOCK_WAIT)
    # Get the apt lock when it's released after do-release-upgrade
    with open("/var/lib/apt/lists/lock", "w") as lockfile:
        fcntl.lockf(lockfile.fileno(), fcntl.LOCK_EX)

    for name in entitlements_enable_order(cfg):
        try:
            entitlement = entitlement_factory(
                cfg=cfg,
                name=name,
                variant="",
            )
        except exceptions.EntitlementNotFoundError:
            LOG.debug("entitlement not found: %s", name)

        application_status, _ = entitlement.application_status()
        applicability_status, _ = entitlement.applicability_status()

        if (
            application_status == ApplicationStatus.ENABLED
            and applicability_status == ApplicabilityStatus.INAPPLICABLE
        ):

            LOG.debug("upgrade-lts-contract disabling %s", name)
            ret = entitlement._perform_disable(progress=ProgressWrapper())

            if not ret:
                LOG.debug("upgrade-lts-contract failed disabling %s", name)

    print(messages.RELEASE_UPGRADE_SUCCESS)


def remove_private_esm_apt_cache():
    system.ensure_folder_absent(defaults.ESM_APT_ROOTDIR)



import json
import logging
import sys
from io import StringIO

import mock
import pytest

from uaclient import log, util

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))
LOG_FMT = "%(asctime)s%(name)s%(funcName)s%(lineno)s\
%(levelname)s%(message)s%(extra)s"
DATE_FMT = "%Y-%m-%dT%H:%M:%S%z"


class TestLogger:
    @pytest.mark.parametrize("caplog_text", [logging.INFO], indirect=True)
    def test_unredacted_text(self, caplog_text):
        text = "Bearer SEKRET"
        LOG.info(text)
        log_text = caplog_text()
        assert text in log_text

    @pytest.mark.parametrize(
        "raw_log,expected",
        (
            ("Super valuable", "Super valuable"),
            (
                "Hi 'Bearer not the droids you are looking for', data",
                "Hi 'Bearer <REDACTED>', data",
            ),
            (
                "Hi 'Bearer not the droids you are looking for', data",
                "Hi 'Bearer <REDACTED>', data",
            ),
            (
                "Executed with sys.argv: ['/usr/bin/ua', 'attach', 'SEKRET']",
                "Executed with sys.argv:"
                " ['/usr/bin/ua', 'attach', '<REDACTED>']",
            ),
            (
                "'resourceTokens': [{'token': 'SEKRET', 'type': 'cc-eal'}]'",
                "'resourceTokens':"
                " [{'token': '<REDACTED>', 'type': 'cc-eal'}]'",
            ),
            (
                "'machineToken': 'SEKRET', 'machineTokenInfo': 'blah'",
                "'machineToken': '<REDACTED>', 'machineTokenInfo': 'blah'",
            ),
            (
                "Failed running command '/usr/lib/apt/apt-helper download-file"
                "https://bearer:S3-Kr3T@esm.ubuntu.com/infra/ubuntu/pool/ "
                "[exit(100)]. Message: Download of file failed"
                " pkgAcquire::Run (13: Permission denied)",
                "Failed running command '/usr/lib/apt/apt-helper download-file"
                "https://bearer:<REDACTED>@esm.ubuntu.com/infra/ubuntu/pool/ "
                "[exit(100)]. Message: Download of file failed"
                " pkgAcquire::Run (13: Permission denied)",
            ),
            (
                "/snap/bin/canonical-livepatch enable S3-Kr3T, foobar",
                "/snap/bin/canonical-livepatch enable <REDACTED> foobar",
            ),
            (
                "Contract value for 'resourceToken' changed to S3kR3T",
                "Contract value for 'resourceToken' changed to <REDACTED>",
            ),
            (
                "data: {'contractToken': 'SEKRET', "
                "'contractTokenInfo':{'expiry'}}",
                "data: {'contractToken': '<REDACTED>', "
                "'contractTokenInfo':{'expiry'}}",
            ),
            (
                "data: {'resourceToken': 'SEKRET', "
                "'entitlement': {'affordances':'blah blah' }}",
                "data: {'resourceToken': '<REDACTED>', "
                "'entitlement': {'affordances':'blah blah' }}",
            ),
            (
                "https://contracts.canonical.com/v1/resources/livepatch"
                "?token=SEKRET: invalid token",
                "https://contracts.canonical.com/v1/resources/livepatch"
                "?token=<REDACTED> invalid token",
            ),
            (
                'data: {"identityToken": "SEket.124-_ys"}',
                'data: {"identityToken": "<REDACTED>"}',
            ),
            (
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: none",
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: none",
            ),
            (
                "response: "
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: none",
                "response: "
                "http://metadata/computeMetadata/v1/instance/service-accounts/"
                "default/identity?audience=contracts.canon, data: <REDACTED>",
            ),
            (
                "'token': 'SEKRET'",
                "'token': '<REDACTED>'",
            ),
            (
                "'userCode': 'SEKRET'",
                "'userCode': '<REDACTED>'",
            ),
            (
                "'magic_token=SEKRET'",
                "'magic_token=<REDACTED>'",
            ),
        ),
    )
    @pytest.mark.parametrize("caplog_text", [logging.INFO], indirect=True)
    def test_redacted_text(self, caplog_text, raw_log, expected):
        LOG.addFilter(log.RegexRedactionFilter())
        LOG.info(raw_log)
        log_text = caplog_text()
        assert expected in log_text


class TestLoggerFormatter:
    @pytest.mark.parametrize(
        "message,level,log_fn,levelname,extra",
        (
            ("mIDValue", logging.DEBUG, LOG.debug, "DEBUG", None),
            ("2B||~2B", logging.INFO, LOG.info, "INFO", None),
            (
                "2B||~2B",
                logging.WARNING,
                LOG.warning,
                "WARNING",
                {"key": "value"},
            ),
        ),
    )
    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    def test_valid_json_output(
        self, caplog_text, message, level, log_fn, levelname, extra
    ):
        formatter = log.JsonArrayFormatter(LOG_FMT, DATE_FMT)
        buffer = StringIO()
        sh = logging.StreamHandler(buffer)
        sh.setLevel(level)
        sh.setFormatter(formatter)
        LOG.addHandler(sh)
        log_fn(message, extra={"extra": extra})
        logged_value = buffer.getvalue()
        val = json.loads(logged_value)
        assert val[1] == levelname
        assert val[2] == util.replace_top_level_logger_name(__name__)
        assert val[5] == message
        if extra:
            assert val[6].get("key") == extra.get("key")
        else:
            assert 7 == len(val)


class TestLogHelpers:
    @pytest.mark.parametrize(
        [
            "we_are_currently_root",
            "expected",
        ],
        [
            (True, "cfg_log_file"),
            (False, "user_log_file"),
        ],
    )
    @mock.patch(
        "uaclient.log.get_user_log_file",
        return_value="user_log_file",
    )
    @mock.patch(
        "uaclient.config.UAConfig.log_file",
        new_callable=mock.PropertyMock,
        return_value="cfg_log_file",
    )
    @mock.patch("uaclient.util.we_are_currently_root")
    def test_get_user_or_root_log_file_path(
        self,
        m_we_are_currently_root,
        m_cfg_log_file,
        m_get_user_log_file,
        we_are_currently_root,
        expected,
    ):
        """
        Tests that the correct log_file path is retrieved
        when the user is root and non-root
        """
        m_we_are_currently_root.return_value = we_are_currently_root
        result = log.get_user_or_root_log_file_path()
        # ensure mocks are used properly
        assert m_cfg_log_file.call_count + m_get_user_log_file.call_count == 1
        if we_are_currently_root:
            assert m_cfg_log_file.call_count == 1
        else:
            assert m_get_user_log_file.call_count == 1
        # ensure correct log_file path is returned
        assert expected == result


@mock.patch("uaclient.log.logging.FileHandler")
@mock.patch("uaclient.log.pathlib.Path")
@mock.patch("uaclient.log.logging.getLogger")
class TestSetupCliLogging:
    def test_correct_handlers_added_to_logger(
        self,
        m_getLogger,
        m_path,
        m_FileHandler,
    ):
        fake_logger = mock.MagicMock(
            handlers=[logging.StreamHandler(sys.stderr)]
        )
        m_getLogger.return_value = fake_logger
        fake_file_handler = mock.MagicMock()
        m_FileHandler.return_value = fake_file_handler

        log.setup_cli_logging(11, "fakefile")
        assert len(fake_logger.handlers) == 0  # handlers is cleared
        assert [mock.call(11)] == fake_file_handler.setLevel.call_args_list
        assert [
            mock.call(fake_file_handler)
        ] == fake_logger.addHandler.call_args_list

    def test_log_file_created_if_not_present(
        self, m_getLogger, m_path, m_FileHandler
    ):
        m_path.return_value.exists.return_value = False
        log.setup_cli_logging(logging.INFO, "fakefile")
        assert m_path.return_value.touch.call_args_list == [
            mock.call(mode=0o640)
        ]



#!/usr/bin/env python3

"""
This script is called after running do-release-upgrade in a machine.
See uaclient/upgrade_lts_contract.py for more details.
"""

import logging

from uaclient import config, defaults, http, log, upgrade_lts_contract

if __name__ == "__main__":
    log.setup_cli_logging(logging.DEBUG, defaults.CONFIG_DEFAULTS["log_level"])
    cfg = config.UAConfig()
    log.setup_cli_logging(cfg.log_level, cfg.log_file)
    http.configure_web_proxy(cfg.http_proxy, cfg.https_proxy)
    upgrade_lts_contract.process_contract_delta_after_apt_lock(cfg)
    upgrade_lts_contract.remove_private_esm_apt_cache()



from behave import when

from features.steps.files import when_i_create_file_with_content
from features.steps.shell import when_i_run_command
from features.util import SUT


@when("I disable access to {website}")
def block_access_to_address(context, website):
    when_i_run_command(
        context,
        "sed -i '1i127.0.0.1 {}' /etc/hosts".format(website),
        "with sudo",
    )


@when("I disable any internet connection on the machine")
@when("I disable any internet connection on the `{machine_name}` machine")
def disable_internet_connection(context, machine_name=SUT):
    when_i_run_command(
        context,
        "ufw default deny incoming",
        "with sudo",
        machine_name=machine_name,
    )
    when_i_run_command(
        context,
        "ufw default deny outgoing",
        "with sudo",
        machine_name=machine_name,
    )
    when_i_run_command(
        context,
        "ufw allow from 10.0.0.0/8",
        "with sudo",
        machine_name=machine_name,
    )
    when_i_run_command(
        context,
        "ufw allow from 172.16.0.0/12",
        "with sudo",
        machine_name=machine_name,
    )
    when_i_run_command(
        context,
        "ufw allow from 192.168.0.0/16",
        "with sudo",
        machine_name=machine_name,
    )
    when_i_run_command(
        context,
        "ufw allow out to 10.0.0.0/8",
        "with sudo",
        machine_name=machine_name,
    )
    when_i_run_command(
        context,
        "ufw allow out to 172.16.0.0/12",
        "with sudo",
        machine_name=machine_name,
    )
    when_i_run_command(
        context,
        "ufw allow out to 192.168.0.0/16",
        "with sudo",
        machine_name=machine_name,
    )
    when_i_run_command(
        context, "ufw allow ssh", "with sudo", machine_name=machine_name
    )
    # We expect DNS to be working, but don't really want to set a server up...
    when_i_run_command(
        context, "ufw allow out 53", "with sudo", machine_name=machine_name
    )
    when_i_run_command(
        context,
        "ufw --force enable",
        "with sudo",
        machine_name=machine_name,
    )


@when(
    "I create a response overlay for `{endpoint}` with response code "
    "`{response_code}` and error message `{error_message}`"
)
def create_response_overlay(context, endpoint, response_code, error_message):
    text = f"""
{{
    "https://contracts.canonical.com{endpoint}": [
    {{
        "code": "{response_code}",
        "response": {{
            "error": "{error_message}"
        }}
    }}
    ]
}}
    """
    when_i_create_file_with_content(
        context=context,
        file_path="/tmp/response-overlay.json",
        text=text,
    )



import mock
import pytest

import uaclient.api.u.pro.attach.magic.wait.v1 as api_wait
from uaclient import exceptions
from uaclient.api.u.pro.attach.magic.wait.v1 import (
    MagicAttachWaitOptions,
    _wait,
)


@mock.patch("uaclient.contract.UAContractClient.get_magic_attach_token_info")
class TestMagicAttachWaitV1:
    @mock.patch("time.sleep")
    def test_wait_succeeds(self, m_sleep, m_attach_token_info, FakeConfig):
        magic_token = "test-id"

        m_attach_token_info.side_effect = [
            {},
            {},
            {
                "token": magic_token,
                "expires": "2100-06-09T18:14:55.323733Z",
                "expiresIn": "2000",
                "userCode": "1234",
                "contractToken": "ctoken",
                "contractID": "cid",
            },
        ]

        options = MagicAttachWaitOptions(magic_token=magic_token)
        expected_response = _wait(options, FakeConfig())

        assert expected_response.contract_token == "ctoken"
        assert expected_response.contract_id == "cid"
        assert 3 == m_attach_token_info.call_count
        assert 2 == m_sleep.call_count

    @mock.patch("time.sleep")
    def test_wait_fails(self, m_sleep, m_attach_token_info, FakeConfig):
        magic_token = "test-id"
        m_attach_token_info.side_effect = [
            {},
            exceptions.MagicAttachTokenError(),
        ]
        options = MagicAttachWaitOptions(magic_token=magic_token)

        with pytest.raises(exceptions.MagicAttachTokenError):
            _wait(options, FakeConfig())

    @mock.patch("time.sleep")
    def test_wait_fails_after_maximum_attempts(
        self, m_sleep, m_attach_token_info, FakeConfig
    ):
        magic_token = "test-id"
        m_attach_token_info.side_effect = [{}, {}, {}]

        options = MagicAttachWaitOptions(magic_token=magic_token)

        with pytest.raises(exceptions.MagicAttachTokenError):
            with mock.patch.object(api_wait, "MAXIMUM_ATTEMPTS", 2):
                _wait(options, FakeConfig())

    @mock.patch("time.sleep")
    def test_wait_fails_after_number_of_connectiviry_errors(
        self, m_sleep, m_attach_token_info, FakeConfig
    ):
        magic_token = "test-id"
        m_attach_token_info.side_effect = [
            exceptions.ConnectivityError(url="url", cause="cause"),
            exceptions.ConnectivityError(url="url", cause="cause"),
            exceptions.ConnectivityError(url="url", cause="cause"),
            exceptions.ConnectivityError(url="url", cause="cause"),
        ]

        options = MagicAttachWaitOptions(magic_token=magic_token)

        with pytest.raises(exceptions.ConnectivityError):
            _wait(options, FakeConfig())

        assert 3 == m_sleep.call_count

    @mock.patch("time.sleep")
    def test_wait_succeeds_after_number_of_connectivity_errors(
        self, m_sleep, m_attach_token_info, FakeConfig
    ):
        magic_token = "test-id"
        m_attach_token_info.side_effect = [
            exceptions.ConnectivityError(url="url", cause="cause"),
            exceptions.ConnectivityError(url="url", cause="cause"),
            exceptions.ConnectivityError(url="url", cause="cause"),
            {
                "token": magic_token,
                "expires": "2100-06-09T18:14:55.323733Z",
                "expiresIn": "2000",
                "userCode": "1234",
                "contractToken": "ctoken",
                "contractID": "cid",
            },
        ]

        options = MagicAttachWaitOptions(magic_token=magic_token)
        expected_response = _wait(options, FakeConfig())

        assert expected_response.contract_token == "ctoken"
        assert expected_response.contract_id == "cid"
        assert 4 == m_attach_token_info.call_count
        assert 3 == m_sleep.call_count

    @mock.patch("time.sleep")
    def test_wait_succeeds_after_unavailable_server(
        self, m_sleep, m_attach_token_info, FakeConfig
    ):
        magic_token = "test-id"
        m_attach_token_info.side_effect = [
            exceptions.ConnectivityError(url="url", cause="cause"),
            exceptions.MagicAttachUnavailable(),
            exceptions.MagicAttachUnavailable(),
            {
                "token": magic_token,
                "expires": "2100-06-09T18:14:55.323733Z",
                "expiresIn": "2000",
                "userCode": "1234",
            },
            {
                "token": magic_token,
                "expires": "2100-06-09T18:14:55.323733Z",
                "expiresIn": "2000",
                "userCode": "1234",
                "contractToken": "ctoken",
                "contractID": "cid",
            },
        ]

        options = MagicAttachWaitOptions(magic_token=magic_token)
        expected_response = _wait(options, FakeConfig())

        assert expected_response.contract_token == "ctoken"
        assert expected_response.contract_id == "cid"
        assert 5 == m_attach_token_info.call_count
        assert 4 == m_sleep.call_count
        assert [
            mock.call(10),
            mock.call(30),
            mock.call(30),
            mock.call(10),
        ] == m_sleep.call_args_list



import datetime
from typing import Any, Dict, List, Optional

from uaclient import system, util
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.api.u.pro.security.cves._common.v1 import (
    VulnerabilityParser,
    get_vulnerabilities,
)
from uaclient.apt import get_apt_cache_datetime
from uaclient.config import UAConfig
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    DatetimeDataValue,
    Field,
    FloatDataValue,
    StringDataValue,
    data_dict,
    data_list,
)


class CVEsOptions(DataObject):
    fields = [
        Field(
            "unfixable",
            BoolDataValue,
            False,
            doc="Show only unfixable CVES.",
        ),
        Field(
            "fixable",
            BoolDataValue,
            False,
            doc="Show only fixable CVES.",
        ),
    ]

    def __init__(
        self,
        *,
        unfixable: Optional[bool] = False,
        fixable: Optional[bool] = False
    ):
        self.unfixable = unfixable
        self.fixable = fixable


class CVEAffectedPackage(DataObject):
    fields = [
        Field(
            "name",
            StringDataValue,
            False,
            doc="The CVE name",
        ),
        Field(
            "fix_version",
            StringDataValue,
            False,
            doc="The version that fixes the CVE for the package",
        ),
        Field(
            "fix_status",
            StringDataValue,
            False,
            doc="The status of the CVE fix for the package",
        ),
        Field(
            "fix_origin",
            StringDataValue,
            doc="The pocket where the fix is available from",
        ),
    ]

    def __init__(
        self, name: str, fix_version: str, fix_status: str, fix_origin: str
    ):
        self.name = name
        self.fix_version = fix_version
        self.fix_status = fix_status
        self.fix_origin = fix_origin


class AffectedPackage(DataObject):
    fields = [
        Field(
            "current_version",
            StringDataValue,
            doc="The current version of the package",
        ),
        Field(
            "cves",
            data_list(CVEAffectedPackage),
            doc="The CVE that affects the package",
        ),
    ]

    def __init__(
        self, *, current_version: str, cves: List[CVEAffectedPackage]
    ):
        self.current_version = current_version
        self.cves = cves


class RelatedUSN(DataObject):
    fields = [
        Field(
            "name",
            StringDataValue,
            doc="The USN name",
        ),
        Field(
            "title",
            StringDataValue,
            doc="The USN title",
        ),
    ]

    def __init__(self, name: str, title: str):
        self.name = name
        self.title = title


class CVEInfo(DataObject):
    fields = [
        Field(
            "description",
            StringDataValue,
            doc="The CVE description",
        ),
        Field(
            "published_at",
            DatetimeDataValue,
            doc="The CVE published date",
        ),
        Field(
            "priority",
            StringDataValue,
            doc="The ubuntu priority for the CVE",
        ),
        Field(
            "notes",
            data_list(StringDataValue),
            False,
            doc="A list of notes for the CVE",
        ),
        Field(
            "cvss_score",
            FloatDataValue,
            False,
            doc="The CVE cvss score",
        ),
        Field(
            "cvss_severity",
            StringDataValue,
            False,
            doc="The CVE cvss severity",
        ),
    ]

    def __init__(
        self,
        *,
        description: str,
        published_at: datetime.datetime,
        priority: str,
        notes: Optional[List[str]] = None,
        cvss_score: Optional[float] = None,
        cvss_severity: Optional[str] = None,
        related_usns: Optional[List[RelatedUSN]] = None,
        related_packages: Optional[List[str]] = None
    ):
        self.description = description
        self.published_at = published_at
        self.priority = priority
        self.notes = notes
        self.cvss_score = cvss_score
        self.cvss_severity = cvss_severity
        # These fields do not appear on the Fields list
        # because we want to access them in the CLI, but
        # not output them in the API
        self.related_usns = related_usns
        self.related_packages = related_packages


class CVEsResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "packages",
            data_dict(value_cls=AffectedPackage),
            doc="A dictionary where the keys are installed package names and the values are AffectedPackage objects.",  # noqa
        ),
        Field(
            "cves",
            data_dict(value_cls=CVEInfo),
            doc="A dictionary where the keys are CVE names and the values are CVEInfo objects.",  # noqa
        ),
    ]

    def __init__(
        self,
        *,
        packages: Dict[str, AffectedPackage],
        cves: Dict[str, CVEInfo],
        # These fields do not appear on the Fields list
        # because we want to access them in the CLI, but
        # not output them in the API
        vulnerability_data_published_at: datetime.datetime,
        apt_updated_at: Optional[datetime.datetime] = None
    ):
        self.packages = packages
        self.cves = cves
        self.vulnerability_data_published_at = vulnerability_data_published_at
        self.apt_updated_at = apt_updated_at


class CVEParser(VulnerabilityParser):
    vulnerability_type = "cves"

    def get_package_vulnerabilities(
        self, affected_pkg: Dict[str, Any]
    ) -> Dict[str, Any]:
        return affected_pkg.get(self.vulnerability_type, {})

    def _post_process_vulnerability_info(
        self,
        vulnerability_info: Dict[str, Any],
        vulnerabilities_data: Dict[str, Any],
    ) -> Dict[str, Any]:
        if vulnerability_info.get("related_usns"):
            related_usns = []

            usn_info = vulnerabilities_data.get("security_issues", {}).get(
                "usns", {}
            )

            for related_usn in vulnerability_info["related_usns"]:
                related_usns.append(
                    {
                        "name": related_usn,
                        "title": usn_info.get(related_usn, {}).get(
                            "title", ""
                        ),
                    }
                )

            vulnerability_info["related_usns"] = related_usns

        return vulnerability_info


def cve_status_match_options(cve, options) -> bool:
    is_fixable = cve.get("fix_version") and cve.get("fix_origin")

    if options.unfixable and is_fixable:
        return False
    elif options.fixable and not is_fixable:
        return False
    return True


def cves(
    options: CVEsOptions,
) -> CVEsResult:
    return _cves(options, UAConfig())


def _parse_vulnerabilities(
    options: CVEsOptions,
    vulnerabilities: Dict[str, Any],
    vulnerability_data_published_at: str,
) -> CVEsResult:
    packages = {}
    allowed_cves = set()
    for pkg_name, package_info in sorted(
        vulnerabilities.get("packages", {}).items()
    ):
        pkg_cves = []
        for cve in sorted(
            package_info.get("cves", []), key=lambda cve: cve["name"]
        ):
            if cve_status_match_options(cve, options):
                pkg_cves.append(
                    CVEAffectedPackage(
                        name=cve["name"],
                        fix_version=cve["fix_version"],
                        fix_status=cve["fix_status"],
                        fix_origin=cve["fix_origin"],
                    )
                )
                allowed_cves.add(cve["name"])

        if pkg_cves:
            packages[pkg_name] = AffectedPackage(
                current_version=package_info["current_version"],
                cves=pkg_cves,
            )

    cves = {
        cve_name: CVEInfo(
            description=cve["description"],
            published_at=util.parse_rfc3339_date(cve["published_at"]),
            priority=cve["ubuntu_priority"],
            notes=cve["notes"],
            cvss_score=cve["cvss_score"],
            cvss_severity=cve["cvss_severity"],
            related_usns=[
                RelatedUSN(
                    name=related_usn.get("name", ""),
                    title=related_usn.get("title", ""),
                )
                for related_usn in cve.get("related_usns", [])
            ],
            related_packages=cve.get("related_packages", []),
        )
        for cve_name, cve in sorted(
            vulnerabilities.get("vulnerabilities", {}).items(),
            key=lambda v: v[0],
        )
        if cve_name in allowed_cves
    }

    return CVEsResult(
        packages=packages,
        cves=cves,
        vulnerability_data_published_at=util.parse_rfc3339_date(
            vulnerability_data_published_at
        ),
        apt_updated_at=get_apt_cache_datetime(),
    )


def _cves(
    options: CVEsOptions,
    cfg: UAConfig,
) -> CVEsResult:
    """
    This endpoint shows the CVE vulnerabilities in the system.
    By default, this API will show all CVEs that affect the system.
    """

    # By default we return all affected CVEs. If a user provides
    # both options, we just switch to the default approach
    if options.unfixable and options.fixable:
        options.unfixable = False
        options.fixable = False

    series = system.get_release_info().series

    cve_vulnerabilities_result = get_vulnerabilities(
        parser=CVEParser(),
        cfg=cfg,
        series=series,
    )

    cve_vulnerabilities = cve_vulnerabilities_result.vulnerabilities_info

    return _parse_vulnerabilities(
        options=options,
        vulnerabilities=cve_vulnerabilities,
        vulnerability_data_published_at=cve_vulnerabilities_result.vulnerability_data_published_at,  # noqa
    )


endpoint = APIEndpoint(
    version="v1",
    name="CVEs",
    fn=_cves,
    options_cls=CVEsOptions,
)

_doc = {
    "introduced_in": "35",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.security.cves.v1 import cves, CVEsOptions

options = CVEsOptions()
result = cves(options)
""",  # noqa: E501
    "result_class": CVEsResult,
    "ignore_result_classes": [DataObject],
    "exceptions": [],
    "example_cli": "pro api u.pro.security.cves.v1",
    "example_json": """
{
    "cves": {
      "CVE-2023-5678": {
        "cvss_score": 8.1,
        "cvss_severity": "high",
        "description": "description example",
        "notes": [
          "note example",
        ],
        "priority": "medium",
        "published_at": ".*"
      }
    },
    "packages": {
      "accountsservice": {
        "current_version": "0.6.40-2ubuntu11.6",
        "cves": [
          {
            "fix_origin": "esm-infra",
            "fix_status": "fixed",
            "fix_version": "0.6.40-2ubuntu11.6+esm1",
            "name": "CVE-2023-5678"
          }
        ]
      },
      "libaccountsservice0": {
        "current_version": "0.6.40-2ubuntu11.6",
        "cves": [
          {
            "fix_origin": "esm-infra",
            "fix_status": "fixed",
            "fix_version": "0.6.40-2ubuntu11.6+esm1",
            "name": "CVE-2023-5678"
          }
        ]
      }
    },
}
""",
}



#!/usr/bin/python3

from datetime import datetime, timedelta, timezone

from uaclient import apt, log
from uaclient.apt_news import update_apt_news
from uaclient.config import UAConfig


def main(cfg: UAConfig):
    if not cfg.apt_news:
        return

    last_update = apt.get_apt_cache_datetime()
    one_day_ago = datetime.now(timezone.utc) - timedelta(days=1)
    if last_update is not None and last_update > one_day_ago:
        return

    update_apt_news(cfg)


if __name__ == "__main__":
    log.setup_journald_logging()
    cfg = UAConfig()
    main(cfg)



import enum
from typing import Optional, Tuple

from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    Field,
    IntDataValue,
    StringDataValue,
)
from uaclient.defaults import (
    CONTRACT_EXPIRY_GRACE_PERIOD_DAYS,
    CONTRACT_EXPIRY_PENDING_DAYS,
)
from uaclient.files import machine_token


class IsAttachedResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "is_attached",
            BoolDataValue,
            doc=(
                "True if the machine is attached to an Ubuntu Pro subscription"
            ),
        ),
        Field(
            "contract_status",
            StringDataValue,
            False,
            doc="Status of the Ubuntu Pro subscription",
        ),
        Field(
            "contract_remaining_days",
            IntDataValue,
            doc="Number of days left in the Ubuntu Pro subscription",
        ),
        Field(
            "is_attached_and_contract_valid",
            BoolDataValue,
            doc=(
                "True if the machine is attached to an Ubuntu Pro subscription"
                " and that subscription is not expired"
            ),
        ),
    ]

    def __init__(
        self,
        *,
        is_attached: bool,
        contract_status: Optional[str],
        contract_remaining_days: int,
        is_attached_and_contract_valid: bool
    ):
        self.is_attached = is_attached
        self.contract_status = contract_status
        self.contract_remaining_days = contract_remaining_days
        self.is_attached_and_contract_valid = is_attached_and_contract_valid


@enum.unique
class ContractExpiryStatus(enum.Enum):
    NONE = None
    ACTIVE = "active"
    ACTIVE_EXPIRED_SOON = "active-soon-to-expire"
    EXPIRED_GRACE_PERIOD = "grace-period"
    EXPIRED = "expired"


def _get_contract_expiry_status(
    cfg: UAConfig,
    is_machine_attached: bool,
    remaining_days: Optional[int],
) -> Tuple[ContractExpiryStatus, int]:
    """Return a tuple [ContractExpiryStatus, num_days]"""
    if not is_machine_attached:
        return ContractExpiryStatus.NONE, 0

    grace_period = CONTRACT_EXPIRY_GRACE_PERIOD_DAYS
    pending_expiry = CONTRACT_EXPIRY_PENDING_DAYS

    # if unknown assume the worst
    if remaining_days is None:
        return ContractExpiryStatus.EXPIRED, -grace_period

    if 0 <= remaining_days <= pending_expiry:
        return ContractExpiryStatus.ACTIVE_EXPIRED_SOON, remaining_days
    elif -grace_period <= remaining_days < 0:
        return ContractExpiryStatus.EXPIRED_GRACE_PERIOD, remaining_days
    elif remaining_days < -grace_period:
        return ContractExpiryStatus.EXPIRED, remaining_days
    return ContractExpiryStatus.ACTIVE, remaining_days


def is_attached() -> IsAttachedResult:
    return _is_attached(UAConfig())


def _is_attached(cfg: UAConfig) -> IsAttachedResult:
    """
    This endpoint shows if the machine is attached to a Pro subscription.
    """
    machine_token_file = machine_token.get_machine_token_file(cfg)
    is_machine_attached = bool(machine_token_file.machine_token)

    contract_status, remaining_days = _get_contract_expiry_status(
        cfg, is_machine_attached, machine_token_file.contract_remaining_days
    )

    is_attached_and_contract_valid = True
    if (
        not is_machine_attached
        or contract_status == ContractExpiryStatus.EXPIRED
    ):
        is_attached_and_contract_valid = False

    return IsAttachedResult(
        is_attached=is_machine_attached,
        contract_status=contract_status.value,
        contract_remaining_days=remaining_days,
        is_attached_and_contract_valid=is_attached_and_contract_valid,
    )


endpoint = APIEndpoint(
    version="v1",
    name="IsAttached",
    fn=_is_attached,
    options_cls=None,
)

_doc = {
    "introduced_in": "28",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.pro.status.is_attached.v1 import is_attached

result = is_attached()
""",  # noqa: E501
    "result_class": IsAttachedResult,
    "exceptions": [],
    "example_cli": "pro api u.pro.status.is_attached.v1",
    "example_json": """
{
    "contract_remaining_days": 360,
    "contract_status": "active",
    "is_attached": true,
    "is_attached_and_contract_valid": true
}
""",
    "extra": """
.. tab-item:: Explanation
    :sync: explanation

    The ``contract_status`` field can return 4 different states, they are:

    * **active**: The contract is currently valid.
    * **grace-period**: The contract is in the grace period. This means that
      it is expired, but there are still some days where the contract will be
      valid.
    * **active-soon-to-expire**: The contract is almost expired, but still
      valid.
    * **expired**: The contract is expired and no longer valid.
""",
}



from uaclient import (
    config,
    daemon,
    entitlements,
    event_logger,
    exceptions,
    messages,
    timer,
    util,
)
from uaclient.cli import cli_util
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.disable import perform_disable
from uaclient.cli.parser import HelpCategory
from uaclient.files import machine_token, state_files
from uaclient.timer.update_messaging import update_motd_messages

event = event_logger.get_event_logger()


@cli_util.verify_json_format_args
@cli_util.assert_root
@cli_util.assert_attached()
@cli_util.assert_lock_file("pro detach")
def action_detach(args, *, cfg, **kwargs) -> int:
    """Perform the detach action for this machine.

    @return: 0 on success, 1 otherwise
    """
    ret = _detach(
        cfg, assume_yes=args.assume_yes, json_output=(args.format == "json")
    )
    if ret == 0:
        daemon.start()
        timer.stop()
    event.process_events()
    return ret


def _detach(cfg: config.UAConfig, assume_yes: bool, json_output: bool) -> int:
    """Detach the machine from the active Ubuntu Pro subscription,

    :param cfg: a ``config.UAConfig`` instance
    :param assume_yes: Assume a yes answer to any prompts requested.
         In this case, it means automatically disable any service during
         detach.
    :param json_output: output should be json only

    @return: 0 on success, 1 otherwise
    """
    to_disable = []
    for ent_name in entitlements.entitlements_disable_order(cfg):
        try:
            ent = entitlements.entitlement_factory(
                cfg=cfg,
                name=ent_name,
            )
        except exceptions.EntitlementNotFoundError:
            continue

        # For detach, we should not consider that a service
        # cannot be disabled because of dependent services,
        # since we are going to disable all of them anyway
        ret, _ = ent.can_disable(ignore_dependent_services=True)
        if ret:
            to_disable.append(ent)

    if to_disable:
        event.info(messages.DETACH_WILL_DISABLE.pluralize(len(to_disable)))
        for ent in to_disable:
            event.info("    {}".format(ent.name))
    if not util.prompt_for_confirmation(assume_yes=assume_yes):
        return 1
    for ent in to_disable:
        perform_disable(
            ent,
            cfg,
            json_output=json_output,
            assume_yes=assume_yes,
            update_status=False,
        )

    machine_token_file = machine_token.get_machine_token_file(cfg)
    machine_token_file.delete()
    state_files.delete_state_files()
    update_motd_messages(cfg)
    event.info(messages.DETACH_SUCCESS)
    return 0


detach_command = ProCommand(
    "detach",
    help=messages.CLI_ROOT_DETACH,
    description=messages.CLI_DETACH_DESC,
    action=action_detach,
    help_category=HelpCategory.OTHER,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "--assume-yes",
                    help=messages.CLI_ASSUME_YES.format(command="detach"),
                    action="store_true",
                ),
                ProArgument(
                    "--format",
                    help=messages.CLI_FORMAT_DESC.format(default="cli"),
                    action="store",
                    choices=["cli", "json"],
                    default="cli",
                ),
            ]
        )
    ],
)



"""
This module is responsible for handling all events
that must be raised to the user somehow. The main idea
behind this module is to centralize all events that happens
during the execution of Pro commands and allows us to report
those events in real time or through a machine-readable format.
"""

import enum
import json
import sys
from typing import Any, Dict, List, Optional, Set, Union  # noqa: F401

from uaclient.yaml import safe_dump

JSON_SCHEMA_VERSION = "0.1"
EventFieldErrorType = Optional[Union[str, Dict[str, str]]]
_event_logger = None


def get_event_logger():
    global _event_logger

    if _event_logger is None:
        _event_logger = EventLogger()

    return _event_logger


@enum.unique
class EventLoggerMode(enum.Enum):
    """
    Defines event logger supported modes.
    Currently, we only support the cli and machine-readable mode. On cli mode,
    we will print to stdout/stderr any event that we receive. Otherwise, we
    will store those events and parse them for the specified format.
    """

    CLI = object()
    JSON = object()
    YAML = object()


def format_machine_readable_output(status: Dict[str, Any]) -> Dict[str, Any]:
    from uaclient.util import get_pro_environment

    status["environment_vars"] = [
        {"name": name, "value": value}
        for name, value in sorted(get_pro_environment().items())
    ]

    # We don't need the origin info in the json output
    status.pop("origin", "")

    # In case there is an error during status and the services were
    # not processed
    status.setdefault("services", [])

    # We are redacting every variant information from the status output
    # because we still are not sure on the best way to represent this
    # information on the status machine readable output
    for service in status.get("services", []):
        if "variants" in service:
            service.pop("variants")

    return status


class EventLogger:
    def __init__(self):
        self._error_events = []  # type: List[Dict[str, EventFieldErrorType]]
        self._warning_events = []  # type: List[Dict[str, EventFieldErrorType]]
        self._processed_services = set()  # type: Set[str]
        self._failed_services = set()  # type: Set[str]
        self._needs_reboot = False
        self._command = ""
        self._output_content = {}

        # By default, the event logger will be on CLI mode,
        # printing every event it receives.
        self._event_logger_mode = EventLoggerMode.CLI

    def reset(self):
        """Reset the state of the event logger attributes."""
        self._error_events = []
        self._warning_events = []
        self._processed_services = set()
        self._failed_services = set()
        self._needs_reboot = False
        self._command = ""
        self._output_content = {}
        self._event_logger_mode = EventLoggerMode.CLI

    def set_event_mode(self, event_mode: EventLoggerMode):
        """Set the event logger mode.

        We currently support the CLI, JSON and YAML modes.
        """
        self._event_logger_mode = event_mode

    def set_command(self, command: str):
        """Set the event logger command.

        The command will tell the process_events method which output method
        to use.
        """
        self._command = command

    def set_output_content(self, output_content: Dict):
        """Set the event logger output content.

        The command will tell the process_events method which content
        to use.
        """
        self._output_content = output_content

    def info(self, info_msg: str, file_type=None, end: Optional[str] = None):
        """
        Print the info message if the event logger is on CLI mode.
        """
        if not file_type:
            file_type = sys.stdout

        if self._event_logger_mode == EventLoggerMode.CLI:
            print(info_msg, file=file_type, end=end)

    def _record_dict_event(
        self,
        msg: str,
        service: Optional[str],
        event_dict: List[Dict[str, EventFieldErrorType]],
        code: Optional[str] = None,
        event_type: Optional[str] = None,
        additional_info: Optional[Dict[str, str]] = None,
    ):
        if event_type is None:
            event_type = "service" if service else "system"

        event_entry = {
            "type": event_type,
            "service": service,
            "message": msg,
            "message_code": code,
        }  # type: Dict[str, EventFieldErrorType]

        if additional_info:
            event_entry["additional_info"] = additional_info

        event_dict.append(event_entry)

    def error(
        self,
        error_msg: str,
        error_code: Optional[str] = None,
        service: Optional[str] = None,
        error_type: Optional[str] = None,
        additional_info: Optional[Dict[str, str]] = None,
    ):
        """
        Store an error in the event logger.

        However, the error will only be stored if the event logger
        is not on CLI mode.
        """
        if self._event_logger_mode != EventLoggerMode.CLI:
            self._record_dict_event(
                msg=error_msg,
                service=service,
                event_dict=self._error_events,
                code=error_code,
                event_type=error_type,
                additional_info=additional_info,
            )

    def warning(self, warning_msg: str, service: Optional[str] = None):
        """
        Store a warning in the event logger.

        However, the warning will only be stored if the event logger
        is not on CLI mode.
        """
        if self._event_logger_mode != EventLoggerMode.CLI:
            self._record_dict_event(
                msg=warning_msg,
                service=service,
                event_dict=self._warning_events,
            )

    def service_processed(self, service: str):
        self._processed_services.add(service)

    def services_failed(self, services: List[str]):
        self._failed_services.update(services)

    def service_failed(self, service: str):
        self._failed_services.add(service)

    def needs_reboot(self, reboot_required: bool):
        self._needs_reboot = reboot_required

    def _generate_failed_services(self):
        services_with_error = {
            error["service"]
            for error in self._error_events
            if error["service"]
        }
        return list(set.union(self._failed_services, services_with_error))

    def _process_events_services(self):
        response = {
            "_schema_version": JSON_SCHEMA_VERSION,
            "result": "success" if not self._error_events else "failure",
            "processed_services": sorted(self._processed_services),
            "failed_services": sorted(self._generate_failed_services()),
            "errors": self._error_events,
            "warnings": self._warning_events,
            "needs_reboot": self._needs_reboot,
        }

        from uaclient.util import DatetimeAwareJSONEncoder

        print(
            json.dumps(response, cls=DatetimeAwareJSONEncoder, sort_keys=True)
        )

    def _process_events_status(self):
        output = format_machine_readable_output(self._output_content)
        output["result"] = "success" if not self._error_events else "failure"
        output["errors"] = self._error_events
        output["warnings"] = self._warning_events

        if self._event_logger_mode == EventLoggerMode.JSON:
            from uaclient.util import DatetimeAwareJSONEncoder

            print(
                json.dumps(
                    output, cls=DatetimeAwareJSONEncoder, sort_keys=True
                )
            )
        elif self._event_logger_mode == EventLoggerMode.YAML:
            print(safe_dump(output, default_flow_style=False))

    def process_events(self) -> None:
        """
        Creates a json response based on all of the
        events stored in the event logger.

        The json response will only be created if the event logger
        is not on CLI mode.
        """
        if self._event_logger_mode != EventLoggerMode.CLI:
            if self._command == "status":
                self._process_events_status()
            else:
                self._process_events_services()



from uaclient import secret_manager
from uaclient.api import exceptions
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.config import UAConfig
from uaclient.contract import UAContractClient
from uaclient.data_types import (
    DataObject,
    Field,
    IntDataValue,
    StringDataValue,
)


class MagicAttachInitiateResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "user_code",
            StringDataValue,
            doc=(
                "Code the user will see in the UI when confirming the Magic"
                " Attach"
            ),
        ),
        Field(
            "token",
            StringDataValue,
            doc=(
                "Magic Token that can be used in either"
                " `u.pro.attach.magic.revoke.v1`_ or"
                " `u.pro.attach.magic.wait.v1`_"
            ),
        ),
        Field(
            "expires",
            StringDataValue,
            doc="Timestamp of the Magic Attach process expiration",
        ),
        Field(
            "expires_in",
            IntDataValue,
            doc="Seconds before the Magic Attach process expires",
        ),
    ]

    def __init__(
        self,
        user_code: str,
        token: str,
        expires: str,
        expires_in: int,
    ):
        self.user_code = user_code
        self.token = token
        self.expires = expires
        self.expires_in = expires_in


def initiate() -> MagicAttachInitiateResult:
    return _initiate(UAConfig())


def _initiate(cfg: UAConfig) -> MagicAttachInitiateResult:
    """
    This endpoint initiates the Magic Attach flow, retrieving the User Code to
    confirm the operation and the Token used to proceed.
    """
    contract = UAContractClient(cfg)
    initiate_resp = contract.new_magic_attach_token()
    secret_manager.secrets.add_secret(initiate_resp["token"])
    secret_manager.secrets.add_secret(initiate_resp["userCode"])
    return MagicAttachInitiateResult(
        user_code=initiate_resp["userCode"],
        token=initiate_resp["token"],
        expires=initiate_resp["expires"],
        expires_in=int(initiate_resp["expiresIn"]),
    )


endpoint = APIEndpoint(
    version="v1",
    name="MagicAttachInitiate",
    fn=_initiate,
    options_cls=None,
)

_doc = {
    "introduced_in": "27.11",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.attach.magic.initiate.v1 import initiate

result = initiate()
""",
    "result_class": MagicAttachInitiateResult,
    "exceptions": [
        (
            exceptions.ConnectivityError,
            (
                "Raised if it is not possible to connect to the contracts"
                " service."
            ),
        ),
        (
            exceptions.ContractAPIError,
            (
                "Raised if there is an unexpected error in the contracts"
                " service interaction."
            ),
        ),
        (
            exceptions.MagicAttachUnavailable,
            (
                "Raised if the Magic Attach service is busy or unavailable at"
                " the moment."
            ),
        ),
    ],
    "example_cli": "pro api u.pro.attach.magic.initiate.v1",
    "example_json": """
{
    "user_code":"<UI_code>",
    "token":"<magic_token>",
    "expires": "<yyyy-MM-dd>T<HH:mm:ss>.<TZ>",
    "expires_in": 600
}
""",
}



import mock
import pytest

from uaclient import exceptions, lock
from uaclient.clouds.aws import AWSAutoAttachInstance
from uaclient.clouds.gcp import GCPAutoAttachInstance
from uaclient.daemon.poll_for_pro_license import (
    attempt_auto_attach,
    poll_for_pro_license,
)

M_PATH = "uaclient.daemon.poll_for_pro_license."


time_mock_curr_value = 0


def time_mock_side_effect_increment_by(increment):
    def _time_mock_side_effect():
        global time_mock_curr_value
        time_mock_curr_value += increment
        return time_mock_curr_value

    return _time_mock_side_effect


@mock.patch(M_PATH + "LOG.info")
@mock.patch(M_PATH + "actions.auto_attach")
@mock.patch(M_PATH + "lock.RetryLock")
class TestAttemptAutoAttach:
    def test_success(self, m_spin_lock, m_auto_attach, m_log_info, FakeConfig):
        cfg = FakeConfig()
        cloud = mock.MagicMock()

        with mock.patch.object(lock, "lock_data_file"):
            attempt_auto_attach(cfg, cloud)

        assert [
            mock.call(lock_holder="pro.daemon.attempt_auto_attach")
        ] == m_spin_lock.call_args_list
        assert [mock.call(cfg, cloud)] == m_auto_attach.call_args_list
        assert [
            mock.call("Successful auto attach")
        ] == m_log_info.call_args_list

    @mock.patch(M_PATH + "system.create_file")
    @mock.patch(M_PATH + "lock.clear_lock_file_if_present")
    @mock.patch(M_PATH + "LOG.error")
    def test_exception(
        self,
        m_log_error,
        m_clear_lock,
        m_create_file,
        m_spin_lock,
        m_auto_attach,
        m_log_info,
        FakeConfig,
    ):
        err = Exception()
        m_auto_attach.side_effect = err
        cfg = FakeConfig()
        cloud = mock.MagicMock()

        attempt_auto_attach(cfg, cloud)

        assert [
            mock.call(lock_holder="pro.daemon.attempt_auto_attach")
        ] == m_spin_lock.call_args_list
        assert [mock.call(cfg, cloud)] == m_auto_attach.call_args_list
        assert [mock.call(err)] == m_log_error.call_args_list
        assert [mock.call()] == m_clear_lock.call_args_list
        assert [
            mock.call("creating flag file to trigger retries")
        ] == m_log_info.call_args_list
        assert [
            mock.call("/run/ubuntu-advantage/flags/auto-attach-failed")
        ] == m_create_file.call_args_list


@mock.patch(M_PATH + "LOG.info")
@mock.patch(M_PATH + "time.sleep")
@mock.patch(M_PATH + "time.time")
@mock.patch(M_PATH + "attempt_auto_attach")
@mock.patch(M_PATH + "GCPAutoAttachInstance.is_pro_license_present")
@mock.patch(M_PATH + "GCPAutoAttachInstance.should_poll_for_pro_license")
@mock.patch(M_PATH + "cloud_instance_factory")
@mock.patch(M_PATH + "system.is_current_series_lts")
@mock.patch(M_PATH + "util.is_config_value_true")
class TestPollForProLicense:
    @pytest.mark.parametrize(
        "is_config_value_true,"
        "is_attached,"
        "is_current_series_lts,"
        "cloud_instance,"
        "should_poll,"
        "is_pro_license_present,"
        "cfg_poll_for_pro_licenses,"
        "expected_log_info_calls,"
        "expected_is_pro_license_present_calls,"
        "expected_attempt_auto_attach_calls",
        [
            (
                True,
                None,
                None,
                None,
                None,
                None,
                None,
                [mock.call("Configured to not auto attach, shutting down")],
                [],
                [],
            ),
            (
                False,
                True,
                None,
                None,
                None,
                None,
                None,
                [mock.call("Already attached, shutting down")],
                [],
                [],
            ),
            (
                False,
                False,
                False,
                None,
                None,
                None,
                None,
                [mock.call("Not on LTS, shutting down")],
                [],
                [],
            ),
            (
                False,
                False,
                True,
                exceptions.CloudFactoryNoCloudError(),
                None,
                None,
                None,
                [mock.call("Not on cloud, shutting down")],
                [],
                [],
            ),
            (
                False,
                False,
                True,
                AWSAutoAttachInstance(),
                None,
                None,
                None,
                [mock.call("Not on supported cloud platform, shutting down")],
                [],
                [],
            ),
            (
                False,
                False,
                True,
                GCPAutoAttachInstance(),
                False,
                None,
                None,
                [mock.call("Not on supported instance, shutting down")],
                [],
                [],
            ),
            (
                False,
                False,
                True,
                GCPAutoAttachInstance(),
                True,
                True,
                None,
                [],
                [mock.call(wait_for_change=False)],
                [mock.call(mock.ANY, mock.ANY)],
            ),
            (
                False,
                False,
                True,
                GCPAutoAttachInstance(),
                True,
                exceptions.CancelProLicensePolling(),
                None,
                [mock.call("Cancelling polling")],
                [mock.call(wait_for_change=False)],
                [],
            ),
            (
                False,
                False,
                True,
                GCPAutoAttachInstance(),
                True,
                False,
                False,
                [
                    mock.call(
                        "Configured to not poll for pro license, shutting down"
                    )
                ],
                [mock.call(wait_for_change=False)],
                [],
            ),
            (
                False,
                False,
                True,
                GCPAutoAttachInstance(),
                True,
                False,
                False,
                [
                    mock.call(
                        "Configured to not poll for pro license, shutting down"
                    )
                ],
                [mock.call(wait_for_change=False)],
                [],
            ),
        ],
    )
    def test_before_polling_loop_checks(
        self,
        m_is_config_value_true,
        m_is_current_series_lts,
        m_cloud_instance_factory,
        m_should_poll,
        m_is_pro_license_present,
        m_attempt_auto_attach,
        m_time,
        m_sleep,
        m_log_info,
        is_config_value_true,
        is_attached,
        is_current_series_lts,
        cloud_instance,
        should_poll,
        is_pro_license_present,
        cfg_poll_for_pro_licenses,
        expected_log_info_calls,
        expected_is_pro_license_present_calls,
        expected_attempt_auto_attach_calls,
        FakeConfig,
        fake_machine_token_file,
    ):
        fake_machine_token_file.attached = is_attached
        cfg = FakeConfig()
        cfg.user_config.poll_for_pro_license = cfg_poll_for_pro_licenses

        m_is_config_value_true.return_value = is_config_value_true
        m_is_current_series_lts.return_value = is_current_series_lts
        m_cloud_instance_factory.side_effect = [cloud_instance]
        m_should_poll.return_value = should_poll
        m_is_pro_license_present.side_effect = [is_pro_license_present]

        poll_for_pro_license(cfg)

        assert expected_log_info_calls == m_log_info.call_args_list
        assert (
            expected_is_pro_license_present_calls
            == m_is_pro_license_present.call_args_list
        )
        assert (
            expected_attempt_auto_attach_calls
            == m_attempt_auto_attach.call_args_list
        )

    @pytest.mark.parametrize(
        "is_pro_license_present_side_effect,"
        "time_side_effect,"
        "expected_is_pro_license_present_calls,"
        "expected_attempt_auto_attach_calls,"
        "expected_log_info_calls,"
        "expected_sleep_calls",
        [
            (
                [False, True],
                time_mock_side_effect_increment_by(100),
                [
                    mock.call(wait_for_change=False),
                    mock.call(wait_for_change=True),
                ],
                [mock.call(mock.ANY, mock.ANY)],
                [],
                [],
            ),
            (
                [False, False, False, False, False, True],
                time_mock_side_effect_increment_by(100),
                [
                    mock.call(wait_for_change=False),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                ],
                [mock.call(mock.ANY, mock.ANY)],
                [],
                [],
            ),
            (
                [False, False, True],
                time_mock_side_effect_increment_by(1),
                [
                    mock.call(wait_for_change=False),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                ],
                [mock.call(mock.ANY, mock.ANY)],
                [
                    mock.call(
                        "wait_for_change returned quickly and no pro license"
                        " present. Waiting %d seconds before polling again",
                        123,
                    )
                ],
                [mock.call(123)],
            ),
            (
                [False, False, False, False, False, True],
                time_mock_side_effect_increment_by(1),
                [
                    mock.call(wait_for_change=False),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                ],
                [mock.call(mock.ANY, mock.ANY)],
                [
                    mock.call(mock.ANY, mock.ANY),
                    mock.call(mock.ANY, mock.ANY),
                    mock.call(mock.ANY, mock.ANY),
                    mock.call(mock.ANY, mock.ANY),
                ],
                [
                    mock.call(123),
                    mock.call(123),
                    mock.call(123),
                    mock.call(123),
                ],
            ),
            (
                [
                    False,
                    False,
                    exceptions.DelayProLicensePolling(),
                    False,
                    exceptions.DelayProLicensePolling(),
                    True,
                ],
                time_mock_side_effect_increment_by(100),
                [
                    mock.call(wait_for_change=False),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                ],
                [mock.call(mock.ANY, mock.ANY)],
                [],
                [mock.call(123), mock.call(123)],
            ),
            (
                [False, False, exceptions.CancelProLicensePolling()],
                time_mock_side_effect_increment_by(100),
                [
                    mock.call(wait_for_change=False),
                    mock.call(wait_for_change=True),
                    mock.call(wait_for_change=True),
                ],
                [],
                [mock.call("Cancelling polling")],
                [],
            ),
        ],
    )
    def test_polling_loop(
        self,
        m_is_config_value_true,
        m_is_current_series_lts,
        m_cloud_instance_factory,
        m_should_poll,
        m_is_pro_license_present,
        m_attempt_auto_attach,
        m_time,
        m_sleep,
        m_log_info,
        is_pro_license_present_side_effect,
        time_side_effect,
        expected_is_pro_license_present_calls,
        expected_attempt_auto_attach_calls,
        expected_log_info_calls,
        expected_sleep_calls,
        FakeConfig,
    ):
        cfg = FakeConfig()
        cfg.user_config.poll_for_pro_license = True
        cfg.user_config.polling_error_retry_delay = 123

        m_is_config_value_true.return_value = False
        m_is_current_series_lts.return_value = True
        m_cloud_instance_factory.return_value = GCPAutoAttachInstance()
        m_should_poll.return_value = True
        m_is_pro_license_present.side_effect = (
            is_pro_license_present_side_effect
        )
        m_time.side_effect = time_side_effect

        poll_for_pro_license(cfg)

        assert expected_sleep_calls == m_sleep.call_args_list
        assert expected_log_info_calls == m_log_info.call_args_list
        assert (
            expected_is_pro_license_present_calls
            == m_is_pro_license_present.call_args_list
        )
        assert (
            expected_attempt_auto_attach_calls
            == m_attempt_auto_attach.call_args_list
        )
